# -- import packages: ---------------------------------------------------
import anndata
import logging
import numpy as np
import sklearn.decomposition
import torch

# -- initialize logger: ----------------------------------------------------
logger = logging.getLogger(name=__name__)
logger.setLevel(logging.INFO)

# -- import local dependencies: ------------------------------------------
from ._fate_perturbation_experiment import PerturbationExperimentResult
from ._perturb import perturb

# -- set typing: -----------------------------------------------------------
from typing import Dict, List, Optional


# -- API-facing function: -------------------------------------------------
def perturb_scan_z_range(
    adata: anndata.AnnData,
    model: "scdiffeq.scDiffEq",
    obs_key: str,
    subset_key: str,
    subset_val: str,
    genes: List[str],
    t_sim: torch.Tensor,
    z_range: np.ndarray,
    PCA: Optional[sklearn.decomposition.PCA] = None,
    seed: int = 0,
    gene_id_key: str = "gene_ids",
    use_key: str = "X_scaled",
    replicates: int = 5,
    N: int = 2000,
    save_simulation: bool = False,
    *args,
    **kwargs,
) -> Dict[float, PerturbationExperimentResult]:
    """Perturb a population of cells and simulate the resulting trajectories over a range of target z-score values.

    Parameters
    ----------
    adata : anndata.AnnData
        The AnnData object containing the data to perturb.
    model : scdiffeq.scDiffEq
        The scDiffEq model to perturb.
    obs_key : str
        Key in `adata.obs` to use for observations.
    subset_key : str
        Key in `adata.obs` to subset the data.
    subset_val : str
        Value in `subset_key` to subset the data.
    genes : List[str]
        List of genes to perturb.
    t_sim : torch.Tensor
        Simulation time points.
    z_range : np.ndarray
        Array of target z-score values for the perturbation.
    PCA : optional
        PCA object for dimensionality reduction. Default is None.
    seed : int, optional
        Random seed for reproducibility. Default is 0.
    gene_id_key : str, optional
        Key in `adata.var` for gene identifiers. Default is "gene_ids".
    use_key : str, optional
        Key in `adata.obsm` for the data to use. Default is "X_scaled".
    replicates : int, optional
        Number of replicates for the perturbation. Default is 5.
    N : int, optional
        Number of cells to perturb. Default is 2000.
    save_simulation : bool, optional
        Whether to save the simulation results. Default is False.
    *args :
        Additional arguments.
    **kwargs :
        Additional keyword arguments.

    Returns
    -------
    Dict[float, PerturbationExperimentResult]
        A dictionary where keys are z-score values and values are the results of the perturbation experiment.
    """

    Perturbed = {}
    for z in z_range:
        Perturbed[z] = perturb(
            adata=adata,
            model=model,
            target_value=z,
            seed=seed,
            N=N,
            t_sim=t_sim,
            obs_key=obs_key,
            subset_key=subset_key,
            subset_val=subset_val,
            gene_id_key=gene_id_key,
            genes=genes,
            replicates=replicates,
            PCA=PCA,
            *args,
            **kwargs,
        )
        logger.info(f"Perturbed {len(genes)} genes (Z={z})")
    return Perturbed




import anndata
import ABCParse
import torch


from ._fate_perturbation_experiment import FatePerturbationExperiment

from typing import Optional, List

import tqdm as tqdm_sh
import tqdm.notebook


class FatePerturbationScreen(ABCParse.ABCParse):
    def __init__(
        self,
        seed: int = 0,
        use_key: str = "X_scaled",
        replicates: int = 5,
        N: int = 200,
        time_key: str = "t",
        nb: bool = True,
        disable_tqdm: Optional[bool] = False,
        *args,
        **kwargs,
    ):
        self.__parse__(locals())

        self.Results = {}

    @property
    def genes(self):
        if not hasattr(self, "_genes"):
            self._genes = self._adata.var_names.tolist()
        return self._genes

    def forward(self, gene):

        prtb_expt = FatePerturbationExperiment(
            seed = self._seed,
            use_key = self._use_key,
            replicates = self._replicates,
            N = self._N,
            time_key = self._time_key,
        )
        result = prtb_expt(
            adata=self._adata,
            model=self._model,
            genes=ABCParse.as_list(gene),
            t_sim=self._t_sim,
            subset_key=self._subset_key,
            subset_val=self._subset_val,
            target_value=self._target_value,
            obs_key=self._obs_key,
            PCA=self._PCA,
        )
        self.Results.update({gene: result})

    @property
    def _progress_bar(self):
        if not self._disable_tqdm:
            if self._nb:
                try:
                    return tqdm.notebook.tqdm(self.genes)
                except:
                    return self.genes
            return tqdm_sh.tqdm(self.genes)
        return self.genes
    
    def __call__(
        self,
        adata: anndata.AnnData,
        model,
        obs_key: str,
        t_sim: torch.Tensor,
        target_value: float = 10,
        genes: Optional[List] = None,
        subset_key="Time point",
        subset_val=2,
        PCA: Optional = None,
        *args,
        **kwargs
    ):
        self.__update__(locals())

        for gene in self._progress_bar: # tqdm.notebook.tqdm(self.genes):
            self.forward(gene)

        return self.Results





from lightning import Callback


class GradientPotentialTest(Callback):
    def on_validation_batch_end(
        self, trainer, model, outputs, batch, batch_idx, dl_idx
    ):

        test_loss = outputs["loss"].item()
        print("Test loss: {}".format(test_loss))



# -- import packages: ---------------------------------------------------------
import abc
import ABCParse
import adata_query
import anndata
import autodevice
import lightning
import logging
import numpy as np
import torch

# -- import local dependencies: -----------------------------------------------
from .utils import L2Norm

# -- configure logger: --------------------------------------------------------
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# -- set type hints: ----------------------------------------------------------
from typing import Any, Dict, Optional, Union

# -- base class: --------------------------------------------------------------
class InstantaneousVelocity(ABCParse.ABCParse):
    """Quantify the instantaneous drift/diffusion given a state and model."""

    def __init__(self, silent: bool = False, *args, **kwargs) -> None:
        """Inhereting class should pass the model and device here."""

        self.__parse__(locals())
        self._L2Norm = L2Norm()

    @property
    def _t(self) -> None:
        return None

    @property
    def _DEVICE(self) -> torch.device:
        if isinstance(self._device, str):
            self._device = autodevice.AutoDevice(self._device)
        return self._device

    @property
    def _INPUT_IS_SIMULATED(self) -> bool:
        return "simulated" in self._adata.uns

    @property
    def _USE_KEY(self) -> str:
        """Some added flexibility for adata_sim, which places the target in adata.X"""
        if self._INPUT_IS_SIMULATED:
            try:
                self._use_key = adata_query.locate(self._adata_sim, self._use_key)
            except:
                return "X"
        return self._use_key

    @property
    def _X(self) -> torch.Tensor:
        return adata_query.fetch(
            self._adata, key=self._USE_KEY, torch=True, device=self._DEVICE
        )

    @property
    def _DiffEq(self):
        return self._model.DiffEq.DiffEq.to(self._DEVICE)

    def _format_outputs(self, X_pred: torch.Tensor) -> np.ndarray:
        """Reshape the output matrix and format as numpy array"""
        X_pred = X_pred.squeeze(dim=-1)
        if self._device.type == "cpu":
            return X_pred.detach().numpy()
        return X_pred.detach().cpu().numpy()

    @abc.abstractmethod
    def forward(self, X): ...

    def _key_action_indicator(self, key, keyset) -> str:
        if not key in keyset:
            return "Added"
        return "Updated"

    def _issue_info_message(self, key: str, key_action: str) -> None:
        logger.info(f"{key_action}: adata.obsm['{key}']")            

    def _add_to_adata(self, X_pred: np.ndarray) -> None:
        if not self._silent:
            logger.info(f"{key_action}: adata.obsm['{key}']")

    def _add_to_adata(self, X_pred: np.ndarray) -> None:
        """
        Add the predicted X_{term} to adata.obsm and compute the L2Norm(X_{term}),
        which is added to adata.obs[key].
        """

        key_action = self._key_action_indicator(
            key=self._obsm_key_added, keyset=self._adata.obsm_keys()
        )
        self._adata.obsm[self._obsm_key_added] = X_pred
        self._issue_info_message(key=self._obsm_key_added, key_action=key_action)

        key_action = self._key_action_indicator(
            key=self._obs_key_added, keyset=self._adata.obs_keys()
        )
        self._adata.obs[self._obs_key_added] = self._L2Norm(X_pred)
        self._issue_info_message(key=self._obs_key_added, key_action=key_action)

    def __call__(
        self,
        adata: anndata.AnnData,
        use_key: str,
        inplace: bool = True,
        *args,
        **kwargs,
    ):

        self.__update__(locals())

        X_pred = self._format_outputs(self.forward(self._X))  # np.ndarray

        if self._inplace:
            self._add_to_adata(X_pred)
        else:
            return X_pred


# -- controller class: --------------------------------------------------------
class InstantaneousDrift(InstantaneousVelocity):
    def __init__(
        self,
        model,
        obsm_key_added: str = "X_drift",
        obs_key_added: str = "drift",
        device=autodevice.AutoDevice(),
        silent: bool = False,
        *args,
        **kwargs,
    ) -> None:
        """ """
        super().__init__(silent=silent)

        self.__parse__(locals())

    def forward(self, X: torch.Tensor):
        return self._DiffEq.f(self._t, X)


class InstantaneousDiffusion(InstantaneousVelocity):
    def __init__(
        self,
        model,
        obsm_key_added: str = "X_diffusion",
        obs_key_added: str = "diffusion",
        device=autodevice.AutoDevice(),
        silent: bool = False,
        *args,
        **kwargs,
    ) -> None:
        
        """ """
        
        super().__init__(silent=silent)

        self.__parse__(locals())

    def forward(self, X: torch.Tensor):
        return self._DiffEq.g(self._t, X)


# -- API-facing functions: ----------------------------------------------------
def drift(
    adata: anndata.AnnData,
    model: "sdq.scDiffEq",
    use_key: str = "X_pca",
    obsm_key_added: str = "X_drift",
    obs_key_added: str = "drift",
    device=autodevice.AutoDevice(),
    inplace: bool = True,
    silent: bool = False,
    *args,
    **kwargs,
):
    """Accessed as sdq.tl.drift(adata) or model.drift(adata)"""

    drift = InstantaneousDrift(
        model=model,
        obsm_key_added=obsm_key_added,
        obs_key_added=obs_key_added,
        device=device,
        silent=silent,
    )
    return drift(adata, use_key=use_key, inplace=inplace)


def diffusion(
    adata: anndata.AnnData,
    model: "sdq.scDiffEq",
    use_key: str = "X_pca",
    obsm_key_added: str = "X_diffusion",
    obs_key_added: str = "diffusion",
    device=autodevice.AutoDevice(),
    inplace: bool = True,
    silent: bool = False,
    *args,
    **kwargs,
):
    """Accessed as sdq.tl.diffusion(adata) or model.diffusion(adata)"""

    diffusion = InstantaneousDiffusion(
        model=model,
        obsm_key_added=obsm_key_added,
        obs_key_added=obs_key_added,
        device=device,
        silent=silent,
    )
    return diffusion(adata, use_key=use_key, inplace=inplace)





def normalize_time(adata, t_min, t_max, time_key="t"):

    """Normalize time to a specified range"""
    
    adata.obs[f"_{time_key}"] = t = adata.obs[time_key]
    init_range = t.max() - t.min()
    init_norm = (t - t.min()) / init_range  # identity, if already norm'd [0,1]
    dest_range = t_max - t_min
    dest_norm = (init_norm * dest_range) + t_min
    adata.obs[time_key] = dest_norm


# -- import packages: ---------------------------------------------------------
import lightning
import logging

# -- import local dependencies: -----------------------------------------------
from .. import utils

# -- set type hints: ----------------------------------------------------------
from typing import List, Optional

# -- configure logging: --------------------------------------------------------
logger = logging.getLogger(__name__)


# -- base mix-in cls: ---------------------------------------------------------
class BaseRoutineMixIn(object):
    """Base train/pre-train routine container"""

    @property
    def _TRAINING_FUNCS(self):
        """"""
        return [self.TrainerGenerator, lightning.Trainer]

    def _check_disable_validation(self, trainer_kwargs) -> None:
        """"""
        if self._train_val_split[1] == 0:
            trainer_kwargs.update(
                {
                    "check_val_every_n_epoch": 0,
                    "limit_val_batches": 0.0,
                    "num_sanity_val_steps": 0.0,
                    "val_check_interval": 0.0,
                },
            )

    def _update_trainer_kwargs(
        self, kwargs, ignore=["version", "working_dir", "logger"]
    ) -> None:
        # -- update kwargs with params already passed (this is problematic)
        kwargs.update(self._PARAMS)

        self._TRAINER_KWARGS = {}
        for func in self._TRAINING_FUNCS:
            self._TRAINER_KWARGS.update(
                utils.extract_func_kwargs(
                    func=func,
                    kwargs=kwargs,
                    ignore=ignore,
                )
            )

        self._check_disable_validation(self._TRAINER_KWARGS)

    def _flag_log_dir(self) -> None:
        if hasattr(self, "_csv_logger"):
            logger.debug(f"Logging locally to: {self._csv_logger.log_dir}")

    def _initialize__attrs(self):
        """ """
        for attr in ["PRETRAIN", "TRAIN"]:
            attr_name = f"_{attr}_CONFIG_COUNT"
            if not hasattr(self, attr_name):
                setattr(self, attr_name, 0)


# -- pre-training mix-in: -----------------------------------------------------
class PreTrainMixIn(BaseRoutineMixIn):
    """Container hosting the pre-training config/execution methods"""

    def _configure_pretrain_step(self, kwargs) -> None:
        """"""

        logger.debug("Configuring pretraining step")

        self._update_trainer_kwargs(kwargs)

        STAGE = "pretrain"

        self._initialize__attrs()

        # -- generate a new trainer
        self.pretrainer = self.TrainerGenerator(
            logger=self.logger,
            max_epochs=self._pretrain_epochs,
            stage=STAGE,
            working_dir=self._working_dir,
            version=self.version,
            pretrain_version=self._PRETRAIN_CONFIG_COUNT,
            train_version=self._TRAIN_CONFIG_COUNT,
            callbacks=self._pretrain_callbacks,
            monitor_hardware=self._monitor_hardware,
            **self._TRAINER_KWARGS,
        )

        self.__flag_log_dir()

        self._PRETRAIN_CONFIG_COUNT += 1

    def pretrain(
        self,
        pretrain_epochs: int = 500,
        pretrain_callbacks: Optional[List] = [],
        ckpt_frequency: int = 25,
        save_last_ckpt: bool = True,
        keep_ckpts: int = -1,
        monitor: Optional[str] = None,
        accelerator: Optional[str] = None,
        log_every_n_steps: int = 1,
        reload_dataloaders_every_n_epochs: int = 1,
        devices: Optional[int] = None,
        deterministic: Optional[bool] = False,
        **kwargs,
    ) -> None:
        """Pretrain method.

        Extended description of the pretrain method.

        Args:
            pretrain_epochs (int): Description. **Default**: 500

            pretrain_callbacks (Optional[List]): Description. **Default**: []

            ckpt_frequency (int): Description. **Default**: 25
        """

        self.__update__(locals())

        if self._pretrain_epochs > 0:
            self._configure_pretrain_step(locals())
            self.pretrainer.fit(self.DiffEq, self.LitDataModule)


# -- training mix-in: ---------------------------------------------------------
class TrainMixIn(BaseRoutineMixIn):
    """Container hosting the training config/execution methods"""

    def _configure_train_step(self, kwargs):
        """"""
        logger.debug(f"Configuring training step")

        self._update_trainer_kwargs(kwargs)

        STAGE = "train"

        self._initialize__attrs()

        # -- generate a new trainer
        self.trainer = self.TrainerGenerator(
            logger=self.logger,
            max_epochs=self._train_epochs,
            stage=STAGE,
            working_dir=self._working_dir,
            version=self.version,
            pretrain_version=self._PRETRAIN_CONFIG_COUNT,
            train_version=self._TRAIN_CONFIG_COUNT,
            callbacks=self._train_callbacks,
            **self._TRAINER_KWARGS,
        )

        self._flag_log_dir()

        self._TRAIN_CONFIG_COUNT += 1

    def train(
        self,
        train_epochs: int = 500,
        train_callbacks: Optional[List] = [],
        ckpt_frequency: int = 25,
        save_last_ckpt: bool = True,
        keep_ckpts: int = -1,
        monitor: Optional[str] = None,
        accelerator: Optional[str] = None,
        log_every_n_steps: int = 1,
        reload_dataloaders_every_n_epochs: int = 1,
        devices: Optional[int] = 1,
        deterministic: Optional[bool] = False,
        **kwargs,
    ) -> None:
        """ """

        self.__update__(locals())

        if self._train_epochs > 0:
            self._configure_train_step(locals())
            self.trainer.fit(self.DiffEq, self.LitDataModule)



# -- import packages: ---------------------------------------------------------
import logging
import sys


# -- configure logger: --------------------------------------------------------
def configure_logging(name="scdiffeq", log_file="scdiffeq.log"):
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    # Prevent adding handlers multiple times (important in notebooks!)
    if logger.hasHandlers():
        return logger

    # Formatter
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )

    # File handler (DEBUG+)
    fh = logging.FileHandler(log_file)
    fh.setLevel(logging.DEBUG)
    fh.setFormatter(formatter)

    stream_formatter = logging.Formatter("scDiffEq [%(levelname)s]: %(message)s")

    # Stream handler (INFO+)
    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(logging.INFO)
    ch.setFormatter(stream_formatter)

    logger.addHandler(fh)
    logger.addHandler(ch)

    return logger




import torch

class LoggingLearnableHParams:
    def __init__(self, value, alt=None):
        
        self.value = value
        
        if not isinstance(alt, type(None)):
            setattr(self, "alt", alt)
        
    @property
    def is_int(self):
        return isinstance(self.value, int)
        
    @property
    def is_float(self):
        return isinstance(self.value, float)
    
    @property
    def is_tensor(self):
        return isinstance(self.value, torch.Tensor)
        
    @property
    def is_learnable_parameter(self):
        return isinstance(self.value, torch.nn.Parameter)
        
    
    def __call__(self):
        if (self.is_float) or (self.is_int):
            return self.value
        elif (self.is_learnable_parameter) or (self.is_tensor):
            if hasattr(self, "alt"):
                return self.alt
            return self.value.item()



# -- import packages: ---------------------------------------------------------
import lightning
import logging
import sys

# -- import local dependencies: -----------------------------------------------
from .. import callbacks

# -- configure logger: --------------------------------------------------------
logger = logging.getLogger(__name__)


# -- config cls: --------------------------------------------------------------
class ProgressBarConfig:
    def __init__(self, total_epochs: int):
        self.total_epochs = total_epochs
        self._pbar = None
        self._env = None

    def _detect_env(self):
        """Returns 'colab', 'jupyter', or 'terminal'."""
        try:
            from IPython import get_ipython

            shell = get_ipython().__class__.__name__
            if "google.colab" in sys.modules:
                return "colab"
            elif shell == "ZMQInteractiveShell":
                return "jupyter"
            else:
                return "terminal"
        except (NameError, ImportError):
            return "terminal"

    def _build_pbar(self, total_epochs: int):

        if self.env == "colab":
            return [lightning.pytorch.callbacks.RichProgressBar()]
        elif self.env == "jupyter":
            return [callbacks.BasicProgressBar(total_epochs=self.total_epochs)]
        else:
            return []

    @property
    def env(self):
        if self._env is None:
            self._env = self._detect_env()
            logger.info(f"Detected environment: {self._env}")
        return self._env

    @property
    def pbar(self):
        if self._pbar is None:
            self._pbar = self._build_pbar(total_epochs=self.total_epochs)
        return self._pbar

    @property
    def enable_progress_bar(self):
        if self.env == "jupyter":
            return False
        else:
            return True




# import functions accessed as sdq.tl.<func>: ---------------------------------
from . import utils

# -- pseudotime: --------------------------------------------------------------
from ._bin_pseudotime import bin_pseudotime

# -- kNN: ---------------------------------------------------------------------
from ._knn import kNN
from ._knn_smoothing import kNNSmoothing


# -- simulation: --------------------------------------------------------------
from ._simulation import Simulation, simulate


# -- post-simulation annotation functions: ------------------------------------
from ._annotate_gene_features import GeneCompatibility, annotate_gene_features
from ._annotate_cell_state import annotate_cell_state
from ._annotate_cell_fate import annotate_cell_fate
from ._invert_scaled_gex import invert_scaled_gex


# -- perturbation: ------------------------------------------------------------
from ._fate_perturbation_experiment import FatePerturbationExperiment
from ._fate_perturbation_screen import FatePerturbationScreen
from ._perturb import perturb
from ._perturb_scan_z_range import perturb_scan_z_range


# -- velocity plotting tools: -------------------------------------------------
from ._grid_velocity import GridVelocity
from ._velocity_embedding import VelocityEmbedding
from ._velocity_graph import velocity_graph



# -- general characterization: ------------------------------------------------
from ._drift_diffusion_state_characterization import drift, diffusion
from ._cell_potential import cell_potential, normalize_cell_potential

__all__ = [
    "bin_pseudotime",
    "kNN",
    "kNNSmoothing",
    "Simulation",
    "simulate",
    "GeneCompatibility",
    "annotate_gene_features",
    "annotate_cell_state",
    "annotate_cell_fate",
    "invert_scaled_gex",
    "FatePerturbationExperiment",
    "FatePerturbationScreen",
    "perturb",
    "perturb_scan_z_range",
    "GridVelocity",
    "VelocityEmbedding",
    "velocity_graph",
    "drift",
    "diffusion",
    "cell_potential",
    "normalize_cell_potential",
    "utils",
]

# -- --------------------------------------------------------------------------
# from ._feature_correlation import (
#     FeatureCorrelation,
#     drift_correlated_features,
#     diffusion_correlated_features,
#     potential_correlated_features,
# )

# from ._grouped_expression import GroupedExpression, grouped_expression
# from ._temporal_expression import temporal_expression





# -- import packages: ---------------------------------------------------------
import neural_diffeqs
import torch

# -- import local dependencies: -----------------------------------------------
from . import mix_ins, base

# -- set type hints: ----------------------------------------------------------
from typing import Literal, Optional, Union, List

# -- lightning model cls: -----------------------------------------------------
class LightningODE_VAE_FixedPotential(
    mix_ins.BaseForwardMixIn,
    base.BaseLightningDiffEq,
    mix_ins.PreTrainMixIn,
    mix_ins.PotentialMixIn,
):
    """LightningODE-VAE-FixedPotential"""
    def __init__(
        self,
        data_dim,
        latent_dim: int = 50,
        name: Optional[str] = None,
        mu_hidden: Union[List[int], int] = [2000, 2000],
        mu_activation: Union[str, List[str]] = "LeakyReLU",
        mu_dropout: Union[float, List[float]] = 0.2,
        mu_bias: bool = True,
        mu_output_bias: bool = True,
        mu_n_augment: int = 0,
        train_lr=1e-5,
        pretrain_lr=1e-3,
        pretrain_epochs=100,
        pretrain_optimizer=torch.optim.Adam,
        train_optimizer=torch.optim.RMSprop,
        pretrain_scheduler=None,
        train_scheduler=torch.optim.lr_scheduler.StepLR,
        pretrain_step_size=None,
        train_step_size=10,
        dt: float =0.1,
        adjoint: bool =False,
        backend: str = "auto",
        loading_existing: bool = False,
        *args,
        **kwargs,
    ) -> None:
        """
        LightningODE_VAE_FixedPotential

        Parameters
        ----------
        data_dim : int
            Dimensionality of the input data.
        latent_dim : int, optional
            Dimensionality of the latent space, by default 50.
        name : str, optional
            Name of the model, by default None.
        mu_hidden : Union[List[int], int], optional
            Hidden layer sizes for the neural network, by default [2000, 2000].
        mu_activation : Union[str, List[str]], optional
            Activation function(s) for the neural network, by default 'LeakyReLU'.
        mu_dropout : Union[float, List[float]], optional
            Dropout rate(s) for the neural network, by default 0.2.
        mu_bias : bool, optional
            Whether to use bias in the neural network, by default True.
        mu_output_bias : bool, optional
            Whether to use bias in the output layer of the neural network, by default True.
        mu_n_augment : int, optional
            Number of augmentations for the neural network, by default 0.
        train_lr : float, optional
            Learning rate for training, by default 1e-5.
        pretrain_lr : float, optional
            Learning rate for pretraining, by default 1e-3.
        pretrain_epochs : int, optional
            Number of epochs for pretraining, by default 100.
        pretrain_optimizer : torch.optim.Optimizer, optional
            Optimizer for pretraining, by default torch.optim.Adam.
        train_optimizer : torch.optim.Optimizer, optional
            Optimizer for training, by default torch.optim.RMSprop.
        pretrain_scheduler : torch.optim.lr_scheduler._LRScheduler, optional
            Learning rate scheduler for pretraining, by default None.
        train_scheduler : torch.optim.lr_scheduler._LRScheduler, optional
            Learning rate scheduler for training, by default torch.optim.lr_scheduler.StepLR.
        pretrain_step_size : int, optional
            Step size for the pretraining learning rate scheduler, by default None.
        train_step_size : int, optional
            Step size for the training learning rate scheduler, by default 10.
        dt : float, optional
            Time step for the ODE solver, by default 0.1.
        adjoint : bool, optional
            Whether to use the adjoint method for the ODE solver, by default False.
        backend : str, optional
            Backend for the ODE solver, by default "auto".
        loading_existing : bool, optional
            Whether to load an existing model, by default False.

        Returns
        -------
        None

        Notes
        -----
        This class implements a VAE with fixed potential ODE using PyTorch Lightning.

        Examples
        --------
        >>> model = LightningODE_VAE_FixedPotential(data_dim=100, latent_dim=20, dt=0.05)
        >>> model.fit(data)
        """
        super().__init__()

        name = self._configure_name(name, loading_existing=loading_existing)

        self.save_hyperparameters()

        # -- torch modules: ----------------------------------------------------
        self._configure_torch_modules(func=neural_diffeqs.PotentialODE, kwargs=locals())
        self._configure_lightning_model(kwargs=locals())

    def forward(self, X0, t, **kwargs):
        """Forward step: (0) integrate in latent space"""
        Z0 = self.Encoder(X0)
        Z_hat = self.integrate(Z0=Z0, t=t, dt=self.hparams["dt"], logqp=False, **kwargs)
        return self.Decoder(Z_hat)

    def step(self, batch, batch_idx, stage=None):

        batch = self.process_batch(batch, batch_idx)
        X_hat = self.forward(batch.X0, batch.t)
        sinkhorn_loss = self.compute_sinkhorn_divergence(
            batch.X, X_hat, batch.W, batch.W_hat
        )
        return self.log_sinkhorn_divergence(sinkhorn_loss).sum()

    def pretrain_step(self, batch, batch_idx, stage=None):

        batch = self.process_batch(batch, batch_idx)
        X0_hat = self.Decoder(self.Encoder(batch.X0))
        recon_loss = self.reconstruction_loss(X0_hat, batch.X0).sum()
        self.log("pretrain_rl_mse", recon_loss.item())
        return recon_loss

    def __repr__(self) -> Literal['LightningODE-VAE-FixedPotential']:
        return "LightningODE-VAE-FixedPotential"



# -- import packages: ---------------------------------------------------------
import ABCParse
import adata_query
import anndata
import numpy as np
import pandas as pd
import sklearn.preprocessing


# -- set typing: --------------------------------------------------------------
from typing import Optional, Union, List


# -- Operational class: -------------------------------------------------------
class InvertScalingExpression(ABCParse.ABCParse):
    def __init__(self, scaler_model, *args, **kwargs):

        self.__parse__(locals(), public=[None])

    def _configure_X_gene(self):

        X_gene = adata_query.fetch(self._adata_sim, self._use_key)

        if isinstance(X_gene, pd.DataFrame):
            X_gene = X_gene.values

        self.ndim = X_gene.ndim
        if self.ndim == 2:
            return X_gene[None, :, :]
        elif self.ndim == 3:
            return X_gene

    @property
    def X_gene(self):
        if not hasattr(self, "_X_gene_processed"):
            self._X_gene_processed = self._configure_X_gene()
        return self._X_gene_processed

    def invert(self):
        """Assumes X_gene is a 3-D array of batch x t x gene"""
        return np.stack(
            [self._scaler_model.inverse_transform(batch) for batch in self.X_gene]
        )

    def adjust_negative_values(self):
        self.X_unscaled[np.where(self.X_unscaled < 0)] = 0
        return self.X_unscaled

    @property
    def _PD_KWARGS(self):

        pd_kwargs = {"index": self._adata_sim.obs.index}

        if not hasattr(self, "_gene_ids") and (
            "gene_ids" in self._adata_sim.uns_keys()
        ):
            setattr(self, "_gene_ids", self._adata_sim.uns["gene_ids"])
        if hasattr(self, "_gene_ids"):
            pd_kwargs.update({"columns": self._gene_ids})

        return pd_kwargs

    def _format_outputs(self, use_key, suffix_added):
        """
        gene_ids
        adata_sim
        """

        if self.ndim == 2:
            self.X_gene_inv = self.X_gene_inv.squeeze(0)

        self.X_gene_inv = pd.DataFrame(self.X_gene_inv, **self._PD_KWARGS)
        self._adata_sim.obsm[f"{use_key}_{suffix_added}"] = self.X_gene_inv

        return self.X_gene_inv

    def __call__(
        self,
        adata_sim,
        use_key: str = "X_gene",
        suffix_added: str = "inv",
        return_df: bool = False,
        gene_ids: Optional[Union[pd.Series, pd.Index, List]] = None,
        silent: bool = False,
        *args,
        **kwargs,
    ) -> pd.DataFrame:

        self.__update__(locals(), public=[None])

        self.X_unscaled = self.invert()
        self.X_gene_inv = self.adjust_negative_values()

        self.X_gene_inv = self._format_outputs(
            use_key=use_key, suffix_added=suffix_added
        )

        if return_df:
            return self.X_gene_inv


# -- API-facing function: -----------------------------------------------------
def invert_scaled_gex(
    adata_sim: anndata.AnnData,
    scaler_model: sklearn.preprocessing.StandardScaler,
    use_key: str = "X_gene",
    suffix_added: str = "inv",
    return_df: bool = False,
    gene_ids: Optional[Union[pd.Series, pd.Index, List]] = None,
    *args,
    **kwargs,
) -> Union[pd.DataFrame, None]:

    """
    Parameters
    ----------
    adata_sim: anndata.AnnData
        Simulated AnnData object. Required.

    scaler_model: sklearn.preprocessing.StandardScaler
        Scaler model used during preprocessing. Required.

    use_key: str, default = "X_gene"
        Key to access the scaled, predicted counts.

    suffix_added: str, default = "inv"

    return_df: bool, default = False
        True, if the pd.DataFrame should be returned in addition to updating
        AnnData. If False, only AnnData is Updated.

    gene_ids: Optional[Union[pd.Series, pd.Index, List]], default = None
        Gene values. Will be assigned as the columns of the resulting
        pd.DataFrame.

    Returns
    -------
    None or X_gene_inv: Union[pd.DataFrame, None]. Updates `adata_sim` either way.

    Notes
    -----
    For a more granular approach to inverting scaler gene
    expression, use `ReverseScalingExpression`.
    """

    inverter = InvertScalingExpression(scaler_model)
    X_gene_inv = inverter(
        adata_sim=adata_sim,
        use_key=use_key,
        suffix_added=suffix_added,
        return_df=return_df,
        gene_ids=gene_ids,
        *args,
        **kwargs,
    )

    if return_df:
        return X_gene_inv



# -- import packages: ---------------------------------------------------------
import ABCParse
import anndata
import time
import logging

 # -- import local dependencies: ----------------------------------------------
from . import utils

# -- set type hints: ----------------------------------------------------------
from typing import Dict, Optional

# -- configure logger: --------------------------------------------------------
logger = logging.getLogger(__name__)

# -- API-facing function: -----------------------------------------------------
def velocity_graph(
    adata: anndata.AnnData,
    state_key: str = "X_pca",
    velocity_key: str = "X_drift",
    n_pcs: Optional[int] = None,
    velocity_key_added: str = "velocity",
    split_negative: bool = True,
    silent: bool = False,
    neighbor_kwargs: Dict = {},
    *args,
    **kwargs,
):
    """
    Args:
        adata (anndata.AnnData)
        
        state_key (str): ... **Default** = "velocity"
        
        velocity_key (str): ... **Default** = "X_drift"
        
        velocity_key_added (str): ... **Default** = "velocity"
        
        split_negative (bool): ... **Default** = True
        
        silent (bool): **Default** = False.
        
        neighbor_kwargs (Optional[Dict[str, Any]]): **Default** = {},
        
    
    Returns:
        (None)
    """
    
    start_time = time.time()
    logger.debug("Starting velocity_graph computation...")
    
    utils.scverse_neighbors(adata, silent=silent, **neighbor_kwargs)
    
    init_kw = ABCParse.function_kwargs(utils.ComputeCosines.__init__, locals())
    call_kw = ABCParse.function_kwargs(utils.ComputeCosines.__call__, locals())

    logger.debug("Starting ComputeCosines computation...")
    cc_start = time.time()
    compute_cosines = utils.ComputeCosines(**init_kw)
    compute_cosines(**call_kw)
    cc_end = time.time()
    logger.debug(f"ComputeCosines computation finished in {cc_end - cc_start:.2f} seconds.")
    logger.debug(f"velocity_graph finished in {time.time() - start_time:.2f} seconds.")




# -- import packages: ---------------------------------------------------------
import ABCParse
import anndata
import cell_perturb
import lightning
import logging
import numpy as np
import os
import pandas as pd
import pathlib
import scipy.stats
import torch


# -- initialize logger: -------------------------------------------------------
logger = logging.getLogger(name=__name__)


# -- import local dependencies: -----------------------------------------------
from ._annotate_cell_state import annotate_cell_state
from ._annotate_cell_fate import annotate_cell_fate
from ._simulation import simulate


# -- set typing: --------------------------------------------------------------
from typing import Dict, List, Optional, Tuple


# -- operational class: -------------------------------------------------------
class PerturbationExperimentResult(ABCParse.ABCParse):
    """Container for the results of a perturbation of experiment. Both the control
    and perturbed arms of the experiment are given as input.
    
    Generally the user interacts with this class, but does not instantiate it. Instead,
    it is instantiated through the output of ``FatePerturbationExperiment``.

    Attributes:
        ctrl
        
        prtb
        
        stats
    """
    def __init__(
        self,
        ctrl_result: anndata.AnnData,
        prtb_result: anndata.AnnData,
        *args,
        **kwargs,
    ) -> None:
        """ Initialize the ``PerturbationExperimentResult`` class.
        
        Args:
            ctrl_result (anndata.AnnData) The resulting AnnData object containing the final state of the simulated control, over each replicate.

            prtb_result (anndata.AnnData). The resulting AnnData object containing the final state of the simulated perturbation, over each replicate.
            
        Returns:
            None
        """

        self.__parse__(locals())

    def _count_fates(self, result: pd.DataFrame):

        fate_counts = (
            result.obs.groupby("replicate")["fate"].value_counts().unstack().fillna(0)
        )
        return fate_counts.div(fate_counts.sum(1), axis=0).T

    def _forward(self, result: anndata.AnnData, key: str):
        return self._count_fates(result)

    #         result.columns = [f"{key}.{i}" for i in result.columns.tolist()]

    @property
    def ctrl(self) -> pd.DataFrame:
        if not hasattr(self, "_ctrl"):
            self._ctrl = self._forward(self._ctrl_result, key="ctrl")
        return self._ctrl

    @property
    def prtb(self) -> pd.DataFrame:
        if not hasattr(self, "_prtb"):
            self._prtb = self._forward(self._prtb_result, key="prtb")
        return self._prtb

    @property
    def _fates(self):
        return list(set(self.ctrl.index.tolist()).union(self.prtb.index.tolist()))

    def _zerofill(self, result_df: pd.DataFrame):
        result_t = result_df.copy().T
        for fate in self._fates:
            if not fate in result_t:
                result_t[fate] = 0

        return result_t

    def _compute_lfc(self, ctrl_t, prtb_t, constant: float = 1e-9):
        return (prtb_t + constant).div((ctrl_t + constant)).apply(np.log2)

    def _compute_pvals(self, ctrl_t, prtb_t):
        return pd.Series(
            {
                fate: scipy.stats.ttest_ind(
                    ctrl_t[fate], prtb_t[fate], equal_var=False
                )[1]
                for fate in self._fates
            }
        )

    def _compute_summary_statistics(self):
        
        ctrl_t = self._zerofill(self.ctrl)
        prtb_t = self._zerofill(self.prtb)
        self._lfc = self._compute_lfc(ctrl_t, prtb_t)
        pvals = self._compute_pvals(ctrl_t, prtb_t)
        
        lfc_pvals = pd.DataFrame([self._lfc.mean(), self._lfc.std(), pvals]).T
        lfc_pvals.columns = ["lfc", 'lfc_std', "pval"]

        return lfc_pvals
    
    @property
    def stats(self):
        if not hasattr(self, "_stats"):
            self._stats = self._compute_summary_statistics()
        return self._stats

    def __repr__(self):
        return "PerturbationExperimentResult"


# -- API-facing operational class: --------------------------------------------
class FatePerturbationExperiment(ABCParse.ABCParse):
    """Container class for an expression perturbation experiment, designed to facilitate
    the analysis of gene expression perturbations and their effects on cell fate and state.
    
    Inherits from ABCParse for abstract base class parsing functionality.
    
    Attributes:
        
    """
    def __init__(
        self,
        seed: int = 0,
        use_key: str = "X_scaled",
        replicates: int = 5,
        N: int = 200,
        time_key: str = "t",
        save_simulation: bool = False,
        save_path: Optional[pathlib.Path] = pathlib.Path("./scdiffeq_simulations"),
        *args,
        **kwargs,
    ) -> None:
        """
        Initializes the FatePerturbationExperiment object.
        
        Args:
            seed (int): Seed for random number generation to ensure reproducibility.

            use_key (str): Key to use for the expression data within the AnnData object.

            replicates (int): Number of replicates to consider in the perturbation experiment.

            N (int): The number of cells to simulate.

            time_key (str): Key to access time-related data within the AnnData object.

            *args, **kwargs: Additional arguments and keyword arguments for flexibility and future extensions.
        """
        self.__parse__(locals())

    @property
    def _PERTURBATION_INIT_KWARGS(self) -> Dict:
        """Retrieves the function keyword arguments for initializing the Perturbation object.
        (i.e., ``cell_perturb.Perturbation.__init__``).
        
        Returns:
            Dict: A dictionary of keyword arguments used for Perturbation object initialization.
        """
        return ABCParse.function_kwargs(
            func=cell_perturb.Perturbation.__init__, kwargs=self._PARAMS
        )

    @property
    def _PERTURBATION_CALL_KWARGS(self) -> Dict:
        """Retrieves the function keyword arguments for calling the Perturbation object
        (i.e., ``cell_perturb.Perturbation.__call__``).
        
        Returns:
            Dict: A dictionary of keyword arguments used for calling the Perturbation object.
        """
        return ABCParse.function_kwargs(
            func=cell_perturb.Perturbation.__call__, kwargs=self._PARAMS
        )

    @property
    def adata_prtb(self) -> anndata.AnnData:
        """Lazily loads or generates the AnnData object resulting from perturbation.
        
        Returns:
            adata_prtb (anndata.AnnData): The AnnData object after applying perturbation.
        """
        if not hasattr(self, "_adata_prtb"):
            self._perturbation = cell_perturb.Perturbation(
                **self._PERTURBATION_INIT_KWARGS
            )
            self._adata_prtb = self._perturbation(**self._PERTURBATION_CALL_KWARGS)
        return self._adata_prtb

    def _subset_final_state(self, adata_sim) -> anndata.AnnData:
        """Extracts the subset of the AnnData object corresponding to the final state of simulation.
        
        Args:
            adata_sim (anndata.AnnData): The simulated AnnData object.
        
        Returns:
            adata_final (anndata.AnnData): A subset of the AnnData object at its final state.
        """

        t = adata_sim.obs[self._time_key]
        return adata_sim[t == t.max()].copy()
    
    @property
    def DiffEq(self) -> lightning.LightningModule:
        """Accessor for the differential equation model used in the simulation.

        Returns:
            DiffEq (lightning.LightningModule)
        """
        if isinstance(self._model.DiffEq, lightning.LightningModule):
            return self._model.DiffEq
        elif isinstance(self._model, lightning.LightningModule):
            return self._model    

    def forward(self) -> Tuple[anndata.AnnData, anndata.AnnData]:
        
        
        """
        Executes the perturbation experiment, comparing control and perturbed conditions.
        
        Simulates, subsets the final simulated state, annotates respective replicates, then annotates
        cell states/fates using the given kNN.
        
        Returns:
            [adata_ctrl, adata_prtb] (Tuple[anndata.AnnData, anndata.AnnData]): A tuple containing AnnData objects for control and perturbation experiments, respectively.
        """
        
        adata_sim_prtb = simulate(
            adata=self.adata_prtb,
            diffeq=self.DiffEq,
            use_key="X_pca_prtb",
            t = self._t_sim,
#             time_key = self._time_key,
        )
        adata_sim_ctrl = simulate(
            adata=self.adata_prtb,
            diffeq=self.DiffEq,
            use_key="X_pca_ctrl",
            t = self._t_sim,
#             time_key = self._time_key,
        )
        
        
        if self._save_simulation:
            
            if not self._save_path.exists():
                os.mkdir(self._save_path)
            
            self.adata_sim_prtb = adata_sim_prtb
            self.adata_sim_ctrl = adata_sim_ctrl
            _genes_ = "_".join(self._genes)
            
            del self.adata_sim_prtb.uns['sim_idx']
            del self.adata_sim_ctrl.uns['sim_idx']
            
            self.adata_sim_prtb.write_h5ad(self._save_path.joinpath(f"adata.{_genes_}.prtb.h5ad"))
            self.adata_sim_ctrl.write_h5ad(self._save_path.joinpath(f"adata.{_genes_}.ctrl.h5ad"))
            logger.info(f"Perturbed simulations saved to: {self._save_path}")
            
        
        prtb = self._subset_final_state(adata_sim_prtb)
        ctrl = self._subset_final_state(adata_sim_ctrl)
        
        rep = self.adata_prtb.obs["replicate"].values
        
        prtb.obs["replicate"] = rep
        ctrl.obs["replicate"] = rep
        
        annotate_cell_state(prtb, kNN = self._model.kNN, obs_key = self._obs_key, silent=True)
        annotate_cell_fate(prtb, state_key = self._obs_key, silent=True)

        annotate_cell_state(ctrl, kNN = self._model.kNN, obs_key = self._obs_key, silent=True)
        annotate_cell_fate(ctrl, state_key = self._obs_key, silent=True)
        
        return ctrl, prtb

    def __call__(
        self,
        adata: anndata.AnnData,
        model: "scdiffeq.scDiffEq",
        t_sim: torch.Tensor,
        obs_key: str,
        genes: List[str],
        subset_key: str,
        subset_val: str,
        target_value: float = 10,
        PCA: Optional = None,
        *args,
        **kwargs,
    ):
        """
        Run perturbation screen.
        
        Args:
            adata (anndata.AnnData): adata obj.
            
            
            model ("scdiffeq.scDiffEq"): scDiffEq model.
            
            genes (List[str]): Genes over which screen should be run.
        
            subset_key (str):
            
            subset_val (str):
        
            target_value (float): Z-score value at which perturbation should be set. **Default**: 10
            
            PCA (Optional[sklearn.decomposition.PCA]: PCA model for transforming expression to model input. **Default**: None.
        
        Returns:
            PerturbationExperimentResult
        """
        self.__update__(locals())

        self.ctrl_result, self.prtb_result = self.forward()
        
        return PerturbationExperimentResult(self.ctrl_result, self.prtb_result)





from ._lightning_callbacks_configuration import LightningCallbacksConfiguration
from ._lightning_trainer_configuration import LightningTrainerConfiguration
from ._lightning_model_configuration import LightningModelConfiguration
from ._data_configuration import DataConfiguration, LightningData, TimeConfiguration, configure_time


# -- import packages: ---------------------------------------------------------
import ABCParse
import logging
import pandas as pd
import pathlib
import torch


# -- set typing: --------------------------------------------------------------
from typing import Union, Dict


# -- set up logging: ----------------------------------------------------------
logger = logging.getLogger(__name__)


# -- operational class: -------------------------------------------------------
class Checkpoint(ABCParse.ABCParse):
    def __init__(self, path: Union[pathlib.Path, str], *args, **kwargs) -> None:
        """Instantiates checkpoint object.

        Args:
            path (Union[pathlib.Path, str]): Path to saved checkpoint.
        """
        self.__parse__(locals())

    @property
    def path(self) -> pathlib.Path:
        """
        Returns:
            pathlib.Path
            The path to the checkpoint.
        """
        return pathlib.Path(self._path)

    @property
    def _fname(self) -> str:
        """
        Returns:
            str
            Filename without extension.
        """
        return self.path.name.split(".")[0]

    @property
    def version(self):
        """
        Returns:
            str
            Version of the checkpoint.
        """
        if not hasattr(self, "_version"):
            v, n = self.path.parent.parent.name.split("_")
            self._version = " ".join([v.capitalize(), n])
        return self._version

    @property
    def _PATH_F_HAT_RAW(self):
        """
        Returns:
            pathlib.Path
            Path to the raw F_hat file.
        """
        if not hasattr(self, "_FATE_PREDICTION_METRICS_PATH"):
            base_path = self.path.parent.parent.joinpath("fate_prediction_metrics")
            converted_name = (
                self.path.name.replace("=", "_").replace("-", ".").split(".ckpt")[0]
            )
            self._FATE_PREDICTION_METRICS_PATH = base_path.joinpath(
                f"{converted_name}/F_hat.unfiltered.csv"
            )
        return self._FATE_PREDICTION_METRICS_PATH

    @property
    def F_hat(self):
        """
        Returns:
            pd.DataFrame or None
            DataFrame containing F_hat data if the path exists, otherwise None.
        """
        if not hasattr(self, "_F_hat"):
            if self._PATH_F_HAT_RAW.exists():
                self._F_hat = pd.read_csv(self._PATH_F_HAT_RAW, index_col=0)
                self._F_hat.index = self._F_hat.index.astype(str)
            else:
                logger.warning(f"F_hat path does not exist.")
                self._F_hat = None
        return self._F_hat

    @property
    def epoch(self) -> Union[int, str]:
        """
        Returns:
            Union[int, str]
            Epoch number if not 'last', otherwise 'last'.
        """
        if self._fname != "last":
            return int(self._fname.split("=")[1].split("-")[0])
        return self._fname

    @property
    def state_dict(self) -> Dict[str, "LightningCheckpoint"]:
        """
        Returns:
            Dict[str, "LightningCheckpoint"]
            State dictionary created by PyTorch Lightning.
        """
        if not hasattr(self, "_ckpt"):
            self._state_dict = torch.load(self.path)  # ["state_dict"]
        return self._state_dict

    def __repr__(self) -> str:
        """
        Returns:
            str
            Object description of checkpoint at epoch.
        """
        return f"ckpt epoch: {self.epoch} [{self.version}]"



# -- import packages: ---------------------------------------------------------
import neural_diffeqs
import torch

# -- import local dependencies: -----------------------------------------------
from . import mix_ins
from . import base

# -- set type hints: ----------------------------------------------------------
from typing import List, Literal, Optional, Union


# -- lightning model: ---------------------------------------------------------
class LightningODE_VAE(
    mix_ins.VAEMixIn,
    mix_ins.PreTrainMixIn,
    base.BaseLightningDiffEq,
):
    """LightningODE-VAE"""
    def __init__(
        self,
        data_dim: int,
        latent_dim: int = 50,
        name: Optional[str] = None,
        train_lr: float = 1e-5,
        pretrain_lr: float = 1e-3,
        pretrain_epochs: int = 100,
        pretrain_optimizer: torch.optim.Optimizer = torch.optim.Adam,
        train_optimizer: torch.optim.Optimizer = torch.optim.RMSprop,
        pretrain_scheduler: torch.optim.lr_scheduler._LRScheduler = torch.optim.lr_scheduler.StepLR,
        train_scheduler: torch.optim.lr_scheduler._LRScheduler = torch.optim.lr_scheduler.StepLR,
        pretrain_step_size: int = 200,
        train_step_size: int = 10,
        dt: float = 0.1,
        adjoint: bool = False,
        backend: str = "auto",
        # -- ode params: -------------------------------------------------------
        mu_hidden: Union[List[int], int] = [400, 400, 400],
        mu_activation: Union[str, List[str]] = "LeakyReLU",
        mu_dropout: Union[float, List[float]] = 0.1,
        mu_bias: bool = True,
        mu_output_bias: bool = True,
        mu_n_augment: int = 0,
        sde_type: str = "ito",
        noise_type: str = "general",
        # -- encoder parameters: -----------------------------------------------
        encoder_n_hidden: int = 4,
        encoder_power: float = 2,
        encoder_activation: Union[str, List[str]] = "LeakyReLU",
        encoder_dropout: Union[float, List[float]] = 0.2,
        encoder_bias: bool = True,
        encoder_output_bias: bool = True,
        # -- decoder parameters: -----------------------------------------------
        decoder_n_hidden: int = 4,
        decoder_power: float = 2,
        decoder_activation: Union[str, List[str]] = "LeakyReLU",
        decoder_dropout: Union[float, List[float]] = 0.2,
        decoder_bias: bool = True,
        decoder_output_bias: bool = True,
        loading_existing: bool = False,
        *args,
        **kwargs,
    ) -> None:
        """
        LightningODE_VAE

        Extended description.

        Parameters
        ----------
        data_dim : int
            Dimension of the data.
        latent_dim : int, optional
            Dimension of the latent space. Default is 50.
        name : str, optional
            Name of the model. Default is None.
        train_lr : float, optional
            Learning rate for training. Default is 1e-5.
        pretrain_lr : float, optional
            Learning rate for pretraining. Default is 1e-3.
        pretrain_epochs : int, optional
            Number of epochs for pretraining. Default is 100.
        pretrain_optimizer : torch.optim.Optimizer, optional
            Optimizer for pretraining. Default is torch.optim.Adam.
        train_optimizer : torch.optim.Optimizer, optional
            Optimizer for training. Default is torch.optim.RMSprop.
        pretrain_scheduler : torch.optim.lr_scheduler._LRScheduler, optional
            Learning rate scheduler for pretraining. Default is torch.optim.lr_scheduler.StepLR.
        train_scheduler : torch.optim.lr_scheduler._LRScheduler, optional
            Learning rate scheduler for training. Default is torch.optim.lr_scheduler.StepLR.
        pretrain_step_size : int, optional
            Step size for pretraining scheduler. Default is 200.
        train_step_size : int, optional
            Step size for training scheduler. Default is 10.
        dt : float, optional
            Time step for ODE solver. Default is 0.1.
        adjoint : bool, optional
            Whether to use the adjoint method for ODE solver. Default is False.
        backend : str, optional
            Backend for ODE solver. Default is "auto".
        mu_hidden : Union[List[int], int], optional
            Hidden layer sizes for the ODE function. Default is [400, 400, 400].
        mu_activation : Union[str, List[str]], optional
            Activation function for the ODE function. Default is "LeakyReLU".
        mu_dropout : Union[float, List[float]], optional
            Dropout rate for the ODE function. Default is 0.1.
        mu_bias : bool, optional
            Whether to use bias in the ODE function. Default is True.
        mu_output_bias : bool, optional
            Whether to use output bias in the ODE function. Default is True.
        mu_n_augment : int, optional
            Number of augmentations for the ODE function. Default is 0.
        sde_type : str, optional
            Type of SDE. Default is "ito".
        noise_type : str, optional
            Type of noise for SDE. Default is "general".
        encoder_n_hidden : int, optional
            Number of hidden layers for the encoder. Default is 4.
        encoder_power : float, optional
            Power for the encoder. Default is 2.
        encoder_activation : Union[str, List[str]], optional
            Activation function for the encoder. Default is "LeakyReLU".
        encoder_dropout : Union[float, List[float]], optional
            Dropout rate for the encoder. Default is 0.2.
        encoder_bias : bool, optional
            Whether to use bias in the encoder. Default is True.
        encoder_output_bias : bool, optional
            Whether to use output bias in the encoder. Default is True.
        decoder_n_hidden : int, optional
            Number of hidden layers for the decoder. Default is 4.
        decoder_power : float, optional
            Power for the decoder. Default is 2.
        decoder_activation : Union[str, List[str]], optional
            Activation function for the decoder. Default is "LeakyReLU".
        decoder_dropout : Union[float, List[float]], optional
            Dropout rate for the decoder. Default is 0.2.
        decoder_bias : bool, optional
            Whether to use bias in the decoder. Default is True.
        decoder_output_bias : bool, optional
            Whether to use output bias in the decoder. Default is True.
        loading_existing : bool, optional
            Whether to load an existing model. Default is False.

        Returns
        -------
        None
        """
        super().__init__()

        name = self._configure_name(name, loading_existing=loading_existing)

        self.save_hyperparameters()

        # -- torch modules: ---------------------------------------------------
        self._configure_torch_modules(func=neural_diffeqs.NeuralODE, kwargs=locals())
        self._configure_lightning_model(kwargs=locals())

    def __repr__(self) -> Literal["LightningODE-VAE"]:
        return "LightningODE-VAE"




# -- import packages: ----------------------------------------------------------
from geomloss import SamplesLoss
import autodevice
import ABCParse


# -- API-facing class: ---------------------------------------------------------
class SinkhornDivergence(SamplesLoss, ABCParse.ABCParse):
    def __init__(
        self,
        device=autodevice.AutoDevice(),
        loss="sinkhorn",
        backend="online",
        p=2,
        blur=0.1,
        scaling=0.7,
        debias=True,
        sample_axis=1,
#         reach=None,
#         diameter=None,
#         truncate=5,
#         cost=None,
#         kernel=None,
#         cluster_scale=None,
#         potentials=False,
#         verbose=False,
        **kwargs
    ) -> None:
        super(SinkhornDivergence, self).__init__()

        self.__parse__(locals())



# -- import packages: ---------------------------------------------------------
import anndata
import sklearn.decomposition
import torch

# -- import local dependencies: -----------------------------------------------
from ._fate_perturbation_experiment import (
    FatePerturbationExperiment,
    PerturbationExperimentResult,
)

# -- set typing: --------------------------------------------------------------
from typing import List, Optional


# -- API-facing function: -----------------------------------------------------
def perturb(
    adata: anndata.AnnData,
    model: "scdiffeq.scDiffEq",
    t_sim: torch.Tensor,
    obs_key: str,
    genes: List[str],
    subset_key: str,
    subset_val: str,
    gene_id_key: str = "gene_ids",
    target_value: float = 10,
    PCA: Optional[sklearn.decomposition.PCA] = None,
    seed: int = 0,
    use_key: str = "X_scaled",
    replicates: int = 5,
    N: int = 200,
    save_simulation: bool = False,
    *args,
    **kwargs,
) -> PerturbationExperimentResult:
    """
    Perturb a population of cells and simulate the resulting trajectories.

    Parameters
    ----------
    adata : anndata.AnnData
        The AnnData object containing the data to perturb.
    model : scdiffeq.scDiffEq
        The scDiffEq model to perturb.
    t_sim : torch.Tensor
        Simulation time points.
    obs_key : str
        Key in `adata.obs` to use for observations.
    genes : List[str]
        List of genes to perturb.
    subset_key : str
        Key in `adata.obs` to subset the data.
    subset_val : str
        Value in `subset_key` to subset the data.
    gene_id_key : str, optional
        Key in `adata.var` for gene identifiers. Default is "gene_ids".
    target_value : float, optional
        Target value for the perturbation. Default is 10.
    PCA : optional
        PCA object for dimensionality reduction. Default is None.
    seed : int, optional
        Random seed for reproducibility. Default is 0.
    use_key : str, optional
        Key in `adata.obsm` for the data to use. Default is "X_scaled".
    replicates : int, optional
        Number of replicates for the perturbation. Default is 5.
    N : int, optional
        Number of cells to perturb. Default is 200.
    save_simulation : bool, optional
        Whether to save the simulation results. Default is False.
    *args :
        Additional arguments.
    **kwargs :
        Additional keyword arguments.

    Returns
    -------
    PerturbationExperimentResult
        The result of the perturbation experiment.
    """
    perturbation = FatePerturbationExperiment(
        seed=seed,
        use_key=use_key,
        replicates=replicates,
        N=N,
        save_simulation=save_simulation,
        *args,
        **kwargs,
    )
    return perturbation(
        adata=adata,
        model=model,
        t_sim=t_sim,
        obs_key=obs_key,
        genes=genes,
        gene_id_key=gene_id_key,
        subset_key=subset_key,
        subset_val=subset_val,
        target_value=target_value,
        PCA=PCA,
    )



from ._intermittent_saves import IntermittentSaves
from ._gradient_potential_callback import GradientPotentialTest
from ._testing import Testing
from ._model_logging import ModelLogging
from ._memory_monitor import MemoryMonitor
from ._basic_progress_bar import BasicProgressBar

__all__ = [
    "IntermittentSaves",
    "GradientPotentialTest",
    "Testing",
    "ModelLogging",
    "MemoryMonitor",
    "BasicProgressBar",
]



# -- import packages: ---------------------------------------------------------
import abc
import ABCParse
import lightning
import torch
import torchsde

# -- import local dependencies: -----------------------------------------------
from ._batch_processor import BatchProcessor
from ._sinkhorn_divergence import SinkhornDivergence

# -- set type hints: ----------------------------------------------------------
from typing import Optional

# -- DiffEq class: ------------------------------------------------------------
class BaseLightningDiffEq(lightning.LightningModule):
    
    """BaseLightningDiffEq: Base class for lightning modules with differential equation solving capabilities.

    This class serves as the base for lightning modules that involve solving differential equations.
    It provides methods for configuring the model, integrating initial value problems (IVPs),
    computing Sinkhorn divergence loss, defining custom steps, and implementing LightningModule methods.

    Attributes:
        None

    Methods:
        __init__: Initialize the BaseLightningDiffEq object.
        _update_lit_diffeq_hparams: Update LightningDifferentialEquation hyperparameters.
        _configure_lightning_model: Configure lightning model with optimizer, scheduler, and additional components.
        _configure_torch_modules: Configure Torch modules for the model.
        PRETRAIN: Property indicating if the model is in pre-training mode.
        _INTEGRATOR: Property returning the integrator function based on hyperparameters.
        integrate: Integrate initial value problems (IVPs) using the specified integrator.
        compute_sinkhorn_divergence: Compute Sinkhorn divergence loss.
        forward: Abstract method for performing forward pass.
        step: Abstract method for performing a step in optimization.
        training_step: LightningModule method for training step.
        validation_step: LightningModule method for validation step.
        test_step: LightningModule method for test step.
        predict_step: LightningModule method for prediction step.
        configure_optimizers: Configure optimizers and schedulers for training.
        __repr__: Return the representation of the class.
        _configure_name: Configure the name of the class.

    """

    def __init__(self, *args, **kwargs):
        """
        Initialize the BaseLightningDiffEq object.

        Args:
            *args: Positional arguments.
            **kwargs: Keyword arguments.

        Returns:
            None

        """
        super().__init__()
        
    def _update_lit_diffeq_hparams(self, model_params) -> None:
        """
        Update LightningDifferentialEquation hyperparameters.

        This method updates the hyperparameters of the lightning differential equation model
        based on the provided model parameters.

        Args:
            model_params (dict): Model parameters.

        Returns:
            None

        """
        for key, val in self.hparams.items():
            if key in model_params.keys():
                if val != model_params[key]:
                    self.hparams.update({key: model_params[key]})
        
    # -- setup: ----------------------------------------------------------------
    def _configure_lightning_model(self, kwargs) -> None:
        
        """
        Configure lightning model with optimizer, scheduler, and additional components.

        This method configures the lightning model with optimizer, scheduler, Sinkhorn divergence,
        batch processor, and other components. It assumes there is no pre-train step, which results
        in a single optimizer, scheduler.

        Args:
            kwargs (dict): Additional keyword arguments.

        Returns:
            None

        """
        
        optimizer = self.hparams['train_optimizer']
        scheduler = self.hparams['train_scheduler']

        self._optimizers = [optimizer(self.parameters(), lr=self.hparams['train_lr'])]
        self._schedulers = [
            scheduler(
                optimizer=self._optimizers[0],
                step_size=self.hparams['train_step_size']),
        ]
        sinkhorn_kwargs = ABCParse.function_kwargs(func = SinkhornDivergence, kwargs = kwargs)
        self.sinkhorn_divergence = SinkhornDivergence(**sinkhorn_kwargs)
        self.process_batch = BatchProcessor
        self.COMPLETED_EPOCHS = 0

    def _configure_torch_modules(self, func, kwargs) -> None:
        """Configure Torch modules for the model.

        This method configures Torch modules for the model based on the provided function and arguments.

        Args:
            func: Function for configuring Torch modules.
            kwargs (dict): Additional keyword arguments.

        Returns:
            None

        """
        kwargs['state_size'] = self.hparams['latent_dim']
        self.DiffEq = func(**ABCParse.function_kwargs(func, kwargs))
        
    @property
    def PRETRAIN(self) -> bool:
        """
        Property indicating if the model is in pre-training mode.

        Returns:
            bool: True if in pre-training mode, False otherwise.

        """
        return False
    
    # -- IVP-solving: ---------------------------------------------------------
    @property
    def _INTEGRATOR(self):
        """Property returning the integrator function based on hyperparameters.

        Returns:
            function: Integrator function.

        """
        if self.hparams["adjoint"]:
            return torchsde.sdeint_adjoint
        return torchsde.sdeint

    def integrate(self, Z0, t, dt, logqp, **kwargs) -> torch.Tensor:
        """Integrate initial value problems (IVPs) using the specified integrator.

        This method integrates initial value problems (IVPs) using the specified integrator.

        Args:
            Z0 (torch.Tensor): Initial latent space.
            t (torch.Tensor): Time steps.
            dt (torch.Tensor): Time step size.
            logqp (torch.Tensor): Logarithmic quantities flag.
            **kwargs: Additional keyword arguments.

        Returns:
            Z_hat (torch.Tensor) predicted (integrated) latent space.

        """
        return self._INTEGRATOR(
            sde=self.DiffEq,
            y0=Z0,
            ts=t,
            dt=dt,
            logqp=logqp,
            **kwargs,
        )

    # -- sinkhorn loss: -------------------------------------------------------
    def compute_sinkhorn_divergence(self, X, X_hat, W, W_hat) -> torch.Tensor:
        """Compute Sinkhorn divergence loss.

        This method computes the Sinkhorn divergence loss between the actual and predicted outputs.

        Args:
            X (tenstorch.Tensoror): Actual outputs.
            X_hat (torch.Tensor): Predicted outputs.
            W (torch.Tensor): Actual weights.
            W_hat (torch.Tensor): Predicted weights.

        Returns:
            tensor: Computed Sinkhorn divergence loss.

        """
        return self.sinkhorn_divergence(
            W_hat.contiguous(), X_hat.contiguous(), W.contiguous(), X.contiguous(), 
        ).requires_grad_()


    # -- custom steps: -------------------------------------------------------------
    @abc.abstractmethod
    def forward(self, Z0, t, **kwargs) -> None:
        """
        Abstract method for performing forward pass. Over-written by inheriting class

        Args:
            Z0 (tensor): Initial latent space.
            t (tensor): Time steps.
            **kwargs: Additional keyword arguments.

        Returns:
            None

        """
        ...

    @abc.abstractmethod
    def step(self, batch, batch_idx, stage=None) -> None:
        """Abstract method for performing a step in optimization.

        Args:
            batch (Batch): Input batch of data.
            batch_idx (int): Index of the batch.
            stage (str): Optional stage information.

        Returns:
            None

        """
        print("WARNING: The base (empty) step is being called from `_base_lightning_diffeq.py`")
        ...

    # -- LightningModule methods: ----------------------------------------------
    def training_step(self, batch, batch_idx, *args, **kwargs) -> None:
        """LightningModule method for training step.

        Args:
            batch (Batch): Input batch of data.
            batch_idx (int): Index of the batch.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            None

        """
        return self.step(batch, batch_idx, stage="training")

    def validation_step(self, batch, batch_idx=None, *args, **kwargs) -> None:
        """LightningModule method for validation step.

        Args:
            batch (Batch): Input batch of data.
            batch_idx (int): Index of the batch.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            None

        """
        return self.step(batch, batch_idx, stage="validation")

    def test_step(self, batch, batch_idx=None, *args, **kwargs) -> None:
        """LightningModule method for test step.

        Args:
            batch (Batch): Input batch of data.
            batch_idx (int): Index of the batch.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            None

        """
        return self.step(batch, batch_idx, stage="test")

    def predict_step(self, batch, batch_idx=None, *args, **kwargs) -> None:
        """LightningModule method for prediction step.

        Args:
            batch (Batch): Input batch of data.
            batch_idx (int): Index of the batch.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            None

        """
        return self.step(batch, batch_idx, stage="predict")

    def configure_optimizers(self):
        """Configure optimizers and schedulers for training.

        Returns:
            tuple: Tuple containing optimizers and schedulers.

        """
        return self._optimizers, self._schedulers
    
    def __repr__(self) -> str:
        """Return the representation of the class.

        Returns:
            str: Representation of the class.

        """
        return "LightningDiffEq"
    
    def _configure_name(self, name: Optional[str] = None, delim: Optional[str] = ".", loading_existing: bool = False) -> str:
        """Configure the name of the class.

        Args:
            name (str, optional): Name of the class.
            delim (str, optional): Delimiter for the name.
            loading_existing (bool, optional): Flag indicating if loading existing class.

        Returns:
            str: Configured name of the class.

        """
                
        if loading_existing or name is None:
            return self.__repr__()
        else:
            return f"{self.__repr__()}{delim}{name}"



# -- import modules: ----------------------------------------------------------
from . import base

# -- import supporting sub-packages: ------------------------------------------
from . import mix_ins

# -- import models: -----------------------------------------------------------
from ._lightning_ode import LightningODE
from ._lightning_sde import LightningSDE
from ._lightning_sde_vae import LightningSDE_VAE
from ._lightning_ode_vae import LightningODE_VAE

from ._lightning_ode_fixed_potential import LightningODE_FixedPotential
from ._lightning_sde_fixed_potential import LightningSDE_FixedPotential
from ._lightning_ode_vae_fixed_potential import LightningODE_VAE_FixedPotential
from ._lightning_sde_vae_fixed_potential import LightningSDE_VAE_FixedPotential


from ._lightning_ode_prior_potential import LightningODE_PriorPotential
from ._lightning_sde_prior_potential import LightningSDE_PriorPotential
from ._lightning_ode_vae_prior_potential import LightningODE_VAE_PriorPotential
from ._lightning_sde_vae_prior_potential import LightningSDE_VAE_PriorPotential

from ._lightning_sde_fate_bias_aware import LightningSDE_FateBiasAware
from ._lightning_sde_vae_fate_bias_aware import LightningSDE_VAE_FateBiasAware


# -- reg velo ratio: ----------------------------------------------------------
from ._lightning_sde_regularized_velocity_ratio import LightningSDE_RegularizedVelocityRatio
from ._lightning_sde_fixed_potential_regularized_velocity_ratio import LightningSDE_FixedPotential_RegularizedVelocityRatio



# -- import packages: ---------------------------------------------------------
import ABCParse
import anndata
import lightning
import logging
import pathlib
import pandas as pd
import torch
import torch_nets
import yaml

# -- import local dependencies: -----------------------------------------------
from ...core import lightning_models, utils, scDiffEq
from ._hparams import HParams
from ._project import Project
from ._checkpoint import Checkpoint
from ._version import Version

# -- set type hints: ----------------------------------------------------------
from typing import Any, Dict, Optional, Union

# -- configure logger: --------------------------------------------------------
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# -- cls: ---------------------------------------------------------------------
class ModelLoader(ABCParse.ABCParse):
    def __init__(
        self,
        project_path=None,
        version=None,
        name_delim: str = ".",
        *args,
        **kwargs,
    ) -> None:
        """ """
        self.__parse__(locals())
        self._validate_version()

    @property
    def project(self) -> Project:
        if not hasattr(self, "_project"):
            self._project = Project(path=self._project_path)
        return self._project

    @property
    def _VERSION_KEY(self) -> str:

        if not hasattr(self, "_VERSION_KEY_PRIVATE"):
            if self._version is None:
                version_key = max(self.project._VERSION_PATHS)
                logger.info(f"Version not provided. Defaulting to: '{version_key}'")
                self._VERSION_KEY_PRIVATE = version_key
            else:
                self._VERSION_KEY_PRIVATE = f"version_{self._version}"

        return self._VERSION_KEY_PRIVATE

    @property
    def _VERSION_PATH(self) -> pathlib.Path:
        return self.project._VERSION_PATHS[self._VERSION_KEY]

    @property
    def version(self) -> Version:
        if not hasattr(self, "_VERSION"):
            self._VERSION = Version(path=self._VERSION_PATH)
        return self._VERSION

    @property
    def hparams(self) -> Dict[str, Any]:
        return self.version.hparams()

    @property
    def _MODEL_TYPE(self):
        name = self.hparams["name"]
        if ":" in name:  # for backwards-compatibility
            return name.split(":")[0].replace("-", "_")
        return name.split(self._name_delim)[0].replace("-", "_")

    @property
    def LightningModule(self):
        return getattr(lightning_models, self._MODEL_TYPE)

    @property
    def ckpt(self):
        return self.version.ckpts[self.epoch]

    @property
    def epoch(self) -> str:
        if not hasattr(self, "_epoch"):
            self._epoch = "last"
            logger.info(f"Epoch not provided. Defaulting to: '{self._epoch}'")
        return self._epoch

    def _validate_epoch(self) -> None:

        msg = f"ckpt at epoch: {self.epoch} does not exist. Choose from: {self.version._SORTED_CKPT_KEYS}"
        assert self.epoch in self.version.ckpts, msg

    def _validate_version(self) -> None:

        version_key = self._VERSION_KEY
        available_versions = [
            attr for attr in self.project.__dir__() if attr.startswith("version_")
        ]

        msg = f"{version_key} not found in project. Available versions: {available_versions}"
        assert version_key in self.project._VERSION_PATHS, msg

    def from_ckpt(self, ckpt_path: str, load_kwargs={"loading_existing": True}):
        return self.LightningModule.load_from_checkpoint(ckpt_path, **load_kwargs)

    def __call__(
        self,
        epoch: Optional[Union[str, int]] = None,
        plot_state_change: Optional[bool] = False,
        load_kwargs={"loading_existing": True},
        *args,
        **kwargs,
    ) -> lightning.LightningModule:
        """
        Parameters
        ----------
        epoch: Optional[Union[str, int]], default = None

        plot_state_change: Optional[bool], default = False

        Returns
        -------
        model:
        """

        self.__update__(locals())

        self._validate_epoch()
        ckpt_path = self.ckpt.path
        return self.LightningModule.load_from_checkpoint(ckpt_path, **load_kwargs)

def _inputs_from_ckpt_path(ckpt_path):
    """If you give the whole ckpt_path, you can derive the other inputs."""

    if isinstance(ckpt_path, str):
        ckpt_path = pathlib.Path(ckpt_path)

    project_path = ckpt_path.parent.parent.parent
    version = int(ckpt_path.parent.parent.name.split("_")[1])
    #     epoch = int(ckpt_path.name.split("-")[0].split("=")[1])

    return {"project_path": project_path, "version": version}  # , "epoch": epoch}


def load_diffeq(
    ckpt_path: Optional[Union[pathlib.Path, str]] = None,
    project_path: Optional[Union[pathlib.Path, str]] = None,
    version: Optional[int] = None,
    epoch: Optional[Union[int, str]] = None,
) -> lightning.LightningModule:
    """
    Load DiffEq from project_path, version [optional], and epoch [optional].

    Parameters
    ----------
    project_path : Union[pathlib.Path, str]
        Path to the project directory.
    version : Optional[int], default=None
        Version number of the model.
    epoch : Optional[Union[int, str]], default=None
        Epoch number of the model.

    Returns
    -------
    lightning.LightningModule
        Lightning differential equation model.
    """

    if not ckpt_path is None:
        inputs = _inputs_from_ckpt_path(ckpt_path)
        project_path = inputs["project_path"]
        version = inputs["version"]
    #         epoch = inputs['epoch']

    model_loader = ModelLoader(project_path=project_path, version=version)

    if not ckpt_path is None:
        return model_loader.from_ckpt(ckpt_path)
    return model_loader(epoch=epoch)


def load_model(
    adata: anndata.AnnData,
    ckpt_path: Optional[Union[pathlib.Path, str]] = None,
    project_path: Optional[Union[pathlib.Path, str]] = None,
    version: Optional[int] = None,
    epoch: Optional[Union[int, str]] = None,
    configure_trainer: Optional[bool] = False,
    #     silent: bool = False,
) -> scDiffEq:
    """
    Load scDiffEq model.

    Parameters
    ----------
    adata : anndata.AnnData
        adata object.
    ckpt_path : Optional[Union[pathlib.Path, str]], optional
        Checkpoint or ckpt_path are accepted, by default None
    project_path : Optional[Union[pathlib.Path, str]], optional
        Project path, by default None
    version : Optional[int], optional
        Version number, by default None
    epoch : Optional[Union[int, str]], optional
        Epoch number, by default None
    configure_trainer : Optional[bool], optional
        Indicate if trainer should be configured, by default False

    Returns
    -------
    scdiffeq.scDiffEq
        Loaded scDiffEq model.
    """

    if isinstance(ckpt_path, Checkpoint):
        ckpt_path = ckpt_path.path

    diffeq = load_diffeq(
        ckpt_path=ckpt_path,
        project_path=project_path,
        version=version,
        epoch=epoch,
    )

    # -- note: if ckpt_path is passed, get: project_path|version|epoch
    if not ckpt_path is None:
        project_path = ckpt_path.parent.parent.parent
        version = int(ckpt_path.parent.parent.name.split("version_")[-1])
        if ckpt_path.name == "last.ckpt":
            epoch = "last"
        else:
            epoch = int(ckpt_path.name.split("epoch=")[-1].split("-")[0])

    # ----------------------------------------------------------------

    sdq_info = {
        "ckpt": ckpt_path.name.split(".ckpt")[0],
        "version": version,
        "project": project_path.name,
    }

    # ----------------------------------------------------------------
    model = scDiffEq(**dict(diffeq.hparams))
    model._load_version = version
    model.configure_data(adata)
    model.configure_kNN()
    model.configure_model(
        diffeq,
        configure_trainer=configure_trainer,
        loading_existing=True,
    )

    PREVIOUS_METRICS_PATH = pathlib.Path(project_path).joinpath(
        f"version_{version}/metrics.csv"
    )
    PREVIOUS_METRICS = pd.read_csv(PREVIOUS_METRICS_PATH)

    model.DiffEq.COMPLETED_EPOCHS = int(PREVIOUS_METRICS["total_epochs"].max())

    model.adata.uns["sdq_info"] = sdq_info

    return model




# -- import packages: ----------------------------------------------------------
import plotly.express as px
import plotly.graph_objects as go
import anndata
import ABCParse

# -- import local dependencies: ------------------------------------------------
from ...core import utils
from ._neighbor_smoothing import NeighborSmoothing
from ._surface_3d import Surface3D
from ._scatter_3d import Scatter3D


# -- set typing: ---------------------------------------------------------------
from typing import Union, Dict

NoneType = type(None)


# -- Operational class: --------------------------------------------------------
class UMAPSurfaceScatterPlot3D(ABCParse.ABCParse):
    """Manager class of both the surface and the scatter"""

    def __init__(
        self,
        use_key="X_umap",
        z_key="psi",
        x_label="UMAP-1",
        y_label="UMAP-2",
        z_label="Ψ, Cell Potential",
        scatter_color="Cell type annotation",
        smoothing_mode: Union[NoneType, str] = "radius",
        radius: int = 1,
        neighbors: int = 15,
        width: float = 800,
        height: float = 800,
        gridpoints: int = 250,
        surface_cmap: str = "aggrnyl",
        scatter_cmap: Union[NoneType, Dict] = None,
        marker_size: int = 2,
        surface_opacity: float = 0.8,
        showscale_surface: bool = False,
        show_surface: bool = True,
        show_scatter: bool = True,
        backgroundcolor = "white",
        gridcolor = "grey",
        showbackground = True,
        zerolinecolor = "white",
        **kwargs,
    ):
        self.__parse__(locals(), public=[None])

    def _fit_neighbor_model(self):
        """Neighbor model is fit to surface data"""
        self.neighbor_model = NeighborSmoothing(
            self._smoothing_mode, self._radius, self._neighbors
        )
        self.neighbor_model.fit(
            self.adata_surface.obsm[self._use_key],
            self.adata_surface.obs[self._z_key].values,
        )

    def _prepare_surface(self):
        surface = Surface3D()
        return surface(self.adata_surface, self.neighbor_model)

    def _prepare_scatter(self):
        scatter = Scatter3D()
        return scatter(self.adata_scatter, self.neighbor_model)

    @property
    def SURFACE(self):
        self.surface_df = self._prepare_surface()
        if not hasattr(self, "_surface"):
            self._surface = go.Surface(
                x=self.surface_df.index,
                y=self.surface_df.columns,
                z=self.surface_df.values,
                opacity=self._surface_opacity,
                colorscale=self._surface_cmap,
                showscale=self._showscale_surface,
            )
        return self._surface

    @property
    def _PLOTLY_SCATTER_KWARGS(self):
        _kw = {"x": "UMAP-1", "y": "UMAP-2", "z": self._z_key}
        if not isinstance(self._scatter_color, NoneType):
            _kw["color"] = self._scatter_color
            self.scatter_df[self._scatter_color] = self.adata_scatter.obs[
                self._scatter_color
            ].values
            if not isinstance(self._scatter_cmap, NoneType):
                _kw["color_discrete_map"] = self._scatter_cmap
        return _kw

    @property
    def SCATTER(self):
        self.scatter_df = self._prepare_scatter()
        if not hasattr(self, "_scatter"):
            self._scatter = px.scatter_3d(
                self.scatter_df, **self._PLOTLY_SCATTER_KWARGS
            )
            self._scatter.update_traces(marker=dict(size=self._marker_size))
        return self._scatter.data

    @property
    def _PLOT_DATA(self):
        _data_to_plot = []
        if self._show_surface:
            _data_to_plot += [self.SURFACE]
        if self._show_scatter:
            _data_to_plot += list(self.SCATTER)
        return _data_to_plot
    
    @property
    def _AXIS_DICT(self):
        return dict(
                    backgroundcolor=self._backgroundcolor,
                    gridcolor=self._gridcolor,
                    showbackground=self._showbackground,
                    zerolinecolor=self._zerolinecolor,
                )

    def plot(self):
        
        fig = go.Figure(data=self._PLOT_DATA)
        fig.update_layout(
            autosize=True,
            width=self._width,
            height=self._height,
            paper_bgcolor="white",
            legend=dict(font=dict(size=14), itemsizing="constant", y=0.2, x=1),
            scene=dict(
                xaxis_title=self._x_label,
                yaxis_title=self._y_label,
                zaxis_title=self._z_label,
                xaxis=self._AXIS_DICT,
                yaxis=self._AXIS_DICT,
                zaxis=self._AXIS_DICT,
            ),
        )
        fig.show()

    def __call__(self, adata_surface, adata_scatter):

        self.__update__(locals())
        self._fit_neighbor_model()
        self.plot()


# -- API-facing function: ------------------------------------------------------
def potential_landscape(
    adata_surface: anndata.AnnData,
    adata_scatter: anndata.AnnData,
    use_key: str = "X_umap",
    z_key: str = "psi",
    x_label: str = "UMAP-1",
    y_label: str = "UMAP-2",
    z_label: str = "Ψ, Cell Potential",
    scatter_color=None,
    smoothing_mode: Union[NoneType, str] = "radius",
    radius: int = 1,
    neighbors: int = 15,
    width: float = 800,
    height: float = 800,
    gridpoints=250,
    surface_cmap: str = "aggrnyl",
    scatter_cmap=None,
    marker_size=2,
    surface_opacity=0.8,
    showscale_surface: bool = False,
    show_surface: bool = True,
    show_scatter: bool = True,
    backgroundcolor = "white",
    gridcolor = "grey",
    showbackground = True,
    zerolinecolor = "white",
):
    """
    Plot the 3D potential landscape and simulated cells within the landscape.

    Parameters:
    -----------
    adata_surface
        type: anndata.AnnData
        default: None

    adata_scatter
        type: anndata.AnnData
        default: None

    use_key
        type: str
        default: "X_umap"

    z_key
        type: str
        default: "psi"

    x_label
        type: str
        default: "UMAP-1"

    y_label
        type: str
        default: "UMAP-2"

    z_label
        type: str
        default: "Ψ, Cell Potential"

    scatter_color
        type: [NoneType, str]
        default: None

    smoothing_mode
        type: [NoneType, str]
        default: "radius"

    radius
        type: float
        default: 1

    neighbors
        type: int
        default: 15

    width
        type: float
        default: 800

    height
        type: float
        default: 800

    gridpoints
        type: int
        default: 250

    surface_cmap
        type: str
        default: "aggrnyl"

    scatter_cmap
        type: [NoneType, Dict]
        default: None

    marker_size
        type: float
        default: 2.0

    surface_opacity
        type: float
        default: 0.8

    showscale_surface
        type: bool
        default: False

    show_surface
        type: bool
        default: True

    show_scatter
        type: bool
        default: True

    Returns:
    --------
    None, displays plotly figure.

    Notes:
    ------
    """
    landscape_plot = UMAPSurfaceScatterPlot3D(
        **utils.function_kwargs(UMAPSurfaceScatterPlot3D, locals())
    )
    landscape_plot(adata_surface, adata_scatter)


# -- configure logging: -------------------------------------------------------
from ._backend_utilities._configure_logging import configure_logging

logger = configure_logging()

# -- handle dependency-related warnings: --------------------------------------
import os as _os

_os.environ["KEOPS_VERBOSE"] = "0"


# -- import model API: --------------------------------------------------------
from .core._scdiffeq import scDiffEq


# -- import sub-packages: -----------------------------------------------------
from . import core
from . import io
from . import plotting as pl
from . import tools as tl
from . import datasets
from . import callbacks
from . import _backend_utilities as utils


# -- version: -----------------------------------------------------------------
from .__version__ import __version__


__all__ = [
    "scDiffEq",
    "callbacks",
    "core",
    "io",
    "pl",
    "tl",
    "datasets",
    "utils",
    "__version__",
    "logger",
]



# -- import packages: ---------------------------------------------------------
import ABCParse
import adata_query
import anndata
import logging
import numpy as np
import warnings

# -- import local dependencies: ----------------------------------------------
from . import utils

# -- set type hints: ----------------------------------------------------------
from typing import Optional

# -- configure logger: --------------------------------------------------------
logger = logging.getLogger(__name__)

# -- operational cls: ---------------------------------------------------------
class VelocityEmbedding(ABCParse.ABCParse):
    """Compute X_emb and V_emb"""

    def __init__(
        self,
        velocity_key: str = "velocity",
        self_transitions: bool = True,
        use_negative_cosines: bool = True,
        T_scale: float = 10,
        *args,
        **kwargs,
    ):
        self.__parse__(locals())

        self._L2Norm = utils.L2Norm()

    @property
    def X_emb(self):
        if not hasattr(self, "_X_emb"):
            self._X_emb = adata_query.fetch(
                self._adata, key=f"X_{self._basis}", torch=False
            )
        return self._X_emb

    @property
    def T(self):
        if not hasattr(self, "_T"):
            self._T = utils.transition_matrix(
                adata=self._adata,
                self_transitions=self._self_transitions,
                use_negative_cosines=self._use_negative_cosines,
                scale=self._T_scale,
                force_dense=False,
                disable_dense=True,
                return_cls=False,
            )
        return self._T

    def forward(self, i):
        # get indices of the ith-cell transitions
        indices = self.T[i].indices
        # compute 2-D coordinate-difference (dX) b/t neighbors
        dX = self.X_emb[indices] - self.X_emb[i, None]  # shape (n_neighbors, 2)

        if not self._retain_scale:
            dX /= self._L2Norm(dX)[:, None]

        # In steady state, dX = 0
        dX[np.isnan(dX)] = 0
        prob = self.T[i].data
        self._V_emb[i] = prob.dot(dX) - prob.mean() * dX.sum(0)

    def __call__(
        self,
        adata: anndata.AnnData,
        basis: str = "umap",
        retain_scale: Optional[bool] = False,
        *args,
        **kwargs,
    ):
        """Compute the velocity embedding (V_emb)

        Args:
            adata (anndata.AnnData)

        Returns:
            V_emb (np.ndarray): Velocity embedding (n_cells, 2)
        """
        self.__update__(locals())

        self._V_emb = np.zeros_like(self.X_emb)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            for i in range(self._adata.n_obs):
                self.forward(i)

        return self.X_emb, self._V_emb




# -- import packages: ---------------------------------------------------------
import ABCParse
import anndata
import numpy as np
import pandas as pd


# -- operational class: -------------------------------------------------------
class PseudotimeBinning(ABCParse.ABCParse):
    def __init__(self, n_bins: int = 10, *args, **kwargs):
        """

        Args:
            n_bins (int): number of non-t0 bins. **Default**: 10.

        """
        self.__parse__(locals())

        self._n_bins = self._n_bins + 2

    @property
    def _BOUNDS(self):
        if not hasattr(self, "_bounds"):
            self._bounds = np.linspace(0, 1, self._n_bins)
        return self._bounds

    @property
    def bins(self):
        if not hasattr(self, "_bins"):
            self._bins = np.array(
                [[i, j] for i, j in zip(self._BOUNDS[:-1], self._BOUNDS[1:])]
            )
        return self._bins

    def _assign_bin(self, value, bins):
        """"""
        match_idx = np.all([value >= self.bins[:, 0], value <= self.bins[:, 1]], axis=0)
        assigned_bin = self.bins[match_idx].flatten()

        return pd.Series(
            {
                "ti": assigned_bin[0],
                "tj": assigned_bin[1],
                "bin": np.where(match_idx)[0][0],
            }
        )

    def _integrate_binned_time(
        self, adata: anndata.AnnData, time_df: pd.DataFrame
    ) -> None:
        """ """
        adata.obs = pd.merge(adata.obs, time_df, how="left", on="index")
        adata.obs["bin"] = adata.obs["bin"].astype(int)
        adata.obs = adata.obs.rename({"bin": "t"}, axis=1)

    @property
    def _TIME_BIN_COLS_PRESENT(self) -> bool:
        return any([key in self._adata.obs for key in ["ti", "tj", "bin", "t"]])

    def __call__(self, adata: anndata.AnnData, pseudotime_key: str):
        """ """

        self.__update__(locals())

        if not self._TIME_BIN_COLS_PRESENT:
            self._time_df = self._adata.obs[self._pseudotime_key].apply(
                self._assign_bin, bins=self.bins
            )
            self._integrate_binned_time(self._adata, self._time_df)


# -- API-facing function: -----------------------------------------------------
def bin_pseudotime(adata: anndata.AnnData, pseudotime_key: str, n_bins: int = 10):
    """
    Args:
        adata (anndata.AnnData)

        pseudotime_key (str) (e.g., "ct_pseudotime")

        n_bins (int) Default = 10

    Returns:
        None
    """
    time_binning = PseudotimeBinning(n_bins=n_bins)
    time_binning(adata=adata, pseudotime_key=pseudotime_key)




import ABCParse
import anndata
import numpy as np
import scipy.sparse
import warnings

warnings.filterwarnings("ignore")

class TransitionMatrix(ABCParse.ABCParse):
    """
    pre-requisite: run `tl.velocity_graph` to compute cosine correlations
    """

    def __init__(
        self,
        self_transitions: bool = True,
        use_negative_cosines: bool = True,
        scale: float = 10,
        *args,
        **kwargs,
    ):
        self.__parse__(locals())

    @property
    def _velocity_graph_key(self):
        return f"{self._velocity_key}_graph"

    @property
    def _velocity_graph_neg_key(self):
        return f"{self._velocity_key}_graph_neg"

    @property
    def _has_velocity_graph(self):
        return self._velocity_graph_key in self._adata.uns

    @property
    def graph(self):
        if not hasattr(self, "_graph"):
            self._graph = self._adata.obsp[self._velocity_graph_key]
        return self._graph

    @property
    def graph_negative(self):
        if not hasattr(self, "_graph_neg"):
            self._graph_neg = self._adata.obsp[self._velocity_graph_neg_key]
        return self._graph_neg

    def compute_self_transitions(self, graph):
        confidence = graph.max(1).A.flatten()
        self_prob = np.clip(np.percentile(confidence, 98) - confidence, 0, 1)
        graph.setdiag(self_prob)
        return graph
    
    def _normalize(self, input):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if scipy.sparse.issparse(input):
                return input.multiply(scipy.sparse.csr_matrix(1.0 / np.abs(input).sum(1)))
            return input / input.sum(1)

    @property
    def auto_densify(self):
        return self._adata.n_obs < 1e4

    def __call__(
        self,
        adata: anndata.AnnData,
        velocity_key: str = "velocity",
        force_dense: bool = False,
        disable_dense: bool = False,
        *args,
        **kwargs,
    ) -> scipy.sparse.csr_matrix:
        """
        Args:
            adata (anndata.AnnData): description.
            
            velocity_key (str): description. **Default** = "velocity"
            
        Returns:
            T (scipy.sparse.csr_matrix): transition matrix.
        """
        
        self.__update__(locals())

        graph = self.compute_self_transitions(self.graph)
        T = np.expm1(graph * self._scale)

        if self._use_negative_cosines:
            T -= np.expm1(-self.graph_negative * self._scale)
        else:
            T += np.expm1(self.graph_negative * self._scale)
            T.data += 1

        T.setdiag(0)
        T.eliminate_zeros()
        
        T = self._normalize(T)
        
        assert sum([self._force_dense, self._disable_dense]) <= 1, "pick one"
        
        if self._force_dense or self.auto_densify and not self._disable_dense:
                T = T.A
        return T


# -- API-facing function: ----------
def transition_matrix(
    adata: anndata.AnnData,
    velocity_key: str = "velocity",
    self_transitions: bool = True,
    use_negative_cosines: bool = True,
    scale: float = 10,
    force_dense: bool = False,
    disable_dense: bool = False,
    return_cls: bool = False,
) -> scipy.sparse.csr_matrix:
    """Compute transition matrix"""
    
    _transition_matrix = TransitionMatrix(
        self_transitions=self_transitions,
        use_negative_cosines=use_negative_cosines,
        scale=scale,
    )
    T = _transition_matrix(adata=adata, velocity_key=velocity_key, force_dense=force_dense, disable_dense=disable_dense)
    
    if return_cls:
        return _transition_matrix, T
    return T


# -- import packages: ---------------------------------------------------------
from neural_diffeqs import LatentPotentialODE
import torch


# -- import local dependencies: -----------------------------------------------
from . import mix_ins
from . import base


# -- set type hints: ----------------------------------------------------------
from typing import Optional, Union, List


# -- lightning model: ---------------------------------------------------------
class LightningODE_VAE_PriorPotential(
    mix_ins.DriftPriorMixIn,
    mix_ins.PotentialMixIn,
    mix_ins.VAEMixIn,
    mix_ins.PreTrainMixIn,
    base.BaseLightningDiffEq,
):
    def __init__(
        self,
        data_dim: int,
        latent_dim: int = 50,
        name: Optional[str] = None,
        train_lr: float = 1e-5,
        pretrain_lr: float = 1e-3,
        pretrain_epochs: int = 100,
        pretrain_optimizer: torch.optim.Optimizer = torch.optim.Adam,
        train_optimizer: torch.optim.Optimizer = torch.optim.RMSprop,
        pretrain_scheduler: torch.optim.lr_scheduler._LRScheduler = torch.optim.lr_scheduler.StepLR,
        train_scheduler: torch.optim.lr_scheduler._LRScheduler = torch.optim.lr_scheduler.StepLR,
        pretrain_step_size: int = 200,
        train_step_size: int = 10,
        dt: float = 0.1,
        adjoint: bool = False,
        backend: str = "auto",
        # -- encoder parameters: -------
        encoder_n_hidden: int = 4,
        encoder_power: float = 2,
        encoder_activation: Union[str, List[str]] = "LeakyReLU",
        encoder_dropout: Union[float, List[float]] = 0.2,
        encoder_bias: bool = True,
        encoder_output_bias: bool = True,
        # -- decoder parameters: -------
        decoder_n_hidden: int = 4,
        decoder_power: float = 2,
        decoder_activation: Union[str, List[str]] = "LeakyReLU",
        decoder_dropout: Union[float, List[float]] = 0.2,
        decoder_bias: bool = True,
        decoder_output_bias: bool = True,
        loading_existing: bool = False,
        *args,
        **kwargs,
    ) -> None:
        """
        LightningODE_VAE_PriorPotential

        Parameters:
        -----------
        data_dim : int
            Dimensionality of the input data
        latent_dim : int, optional
            Dimensionality of the latent space, by default 50
        name : str, optional
            Name of the model, by default None
        train_lr : float, optional
            Learning rate for training, by default 1e-5
        pretrain_lr : float, optional
            Learning rate for pretraining, by default 1e-3
        pretrain_epochs : int, optional
            Number of epochs for pretraining, by default 100
        pretrain_optimizer : torch.optim.Optimizer, optional
            Optimizer for pretraining, by default torch.optim.Adam
        train_optimizer : torch.optim.Optimizer, optional
            Optimizer for training, by default torch.optim.RMSprop
        pretrain_scheduler : torch.optim.lr_scheduler._LRScheduler, optional
            Learning rate scheduler for pretraining, by default torch.optim.lr_scheduler.StepLR
        train_scheduler : torch.optim.lr_scheduler._LRScheduler, optional
            Learning rate scheduler for training, by default torch.optim.lr_scheduler.StepLR
        pretrain_step_size : int, optional
            Step size for the pretraining learning rate scheduler, by default 200
        train_step_size : int, optional
            Step size for the training learning rate scheduler, by default 10
        dt : float, optional
            Time step for the ODE solver, by default 0.1
        adjoint : bool, optional
            Whether to use the adjoint method for the ODE solver, by default False
        backend : str, optional
            Backend for the ODE solver, by default "auto"
        encoder_n_hidden : int, optional
            Number of hidden layers in the encoder, by default 4
        encoder_power : float, optional
            Power of the encoder, by default 2
        encoder_activation : Union[str, List[str]], optional
            Activation function(s) for the encoder, by default 'LeakyReLU'
        encoder_dropout : Union[float, List[float]], optional
            Dropout rate(s) for the encoder, by default 0.2
        encoder_bias : bool, optional
            Whether to use bias in the encoder, by default True
        encoder_output_bias : bool, optional
            Whether to use bias in the output layer of the encoder, by default True
        decoder_n_hidden : int, optional
            Number of hidden layers in the decoder, by default 4
        decoder_power : float, optional
            Power of the decoder, by default 2
        decoder_activation : Union[str, List[str]], optional
            Activation function(s) for the decoder, by default 'LeakyReLU'
        decoder_dropout : Union[float, List[float]], optional
            Dropout rate(s) for the decoder, by default 0.2
        decoder_bias : bool, optional
            Whether to use bias in the decoder, by default True
        decoder_output_bias : bool, optional
            Whether to use bias in the output layer of the decoder, by default True
        loading_existing : bool, optional
            Whether to load an existing model, by default False

        Returns:
        --------
        None

        Notes:
        ------
        This class implements a VAE with prior potential ODE using PyTorch Lightning.

        Examples:
        ---------
        >>> model = LightningODE_VAE_PriorPotential(data_dim=100, latent_dim=20, dt=0.05)
        >>> model.fit(data)
        """
        super().__init__()

        name = self._configure_name(name, loading_existing=loading_existing)

        self.save_hyperparameters()

        # -- torch modules: ----------------------------------------------------
        self._configure_torch_modules(func=LatentPotentialODE, kwargs=locals())
        self._configure_lightning_model(kwargs=locals())

    def log_computed_loss(self, sinkhorn_loss, t, kl_div_loss, stage):

        sinkhorn_loss = self.log_sinkhorn_divergence(sinkhorn_loss).sum()
        self.log(f"kl_div_{stage}", kl_div_loss.sum())

        return sinkhorn_loss + kl_div_loss.sum()

    def step(self, batch, batch_idx, stage=None):

        batch = self.process_batch(batch, batch_idx)
        X_hat, kl_div_loss = self.forward(batch.X0, batch.t)
        sinkhorn_loss = self.compute_sinkhorn_divergence(
            batch.X, X_hat, batch.W, batch.W_hat
        )
        return self.log_computed_loss(
            sinkhorn_loss, t=batch.t, kl_div_loss=kl_div_loss, stage=stage
        )

    def pretrain_step(self, batch, batch_idx, stage=None):

        batch = self.process_batch(batch, batch_idx)
        X0_hat = self.Decoder(self.Encoder(batch.X0))
        recon_loss = self.reconstruction_loss(X0_hat, batch.X0).sum()
        self.log("pretrain_rl_mse", recon_loss.item())
        return recon_loss

    def __repr__(self):
        return "LightningODE-VAE-PriorPotential"



# -- import packages: ---------------------------------------------------------
import requests
import tqdm


# -- operational cls: ---------------------------------------------------------
class FigshareDownloader:
    def __init__(
        self,
        chunk_size: int = int(8 * 1024 * 1024),
    ):
        """
        Args:
            chunk_size: int
                The size of the chunk to download the file in.
        """
        self.chunk_size = chunk_size

    @property
    def url(self):
        return f"https://figshare.com/ndownloader/files/{self.figshare_id}"

    def _get_size(self, response):
        return int(response.headers.get("Content-Length", 0))

    def download(self, response, write_path: str):

        total_size = self._get_size(response)

        with open(write_path, "wb") as f:
            with tqdm.tqdm(
                total=total_size,
                unit="iB",
                unit_scale=True,
                desc="Downloading",
                ncols=100,
                bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt} [{elapsed}<{remaining}]",
            ) as pbar:
                for chunk in response.iter_content(chunk_size=self.chunk_size):
                    size = f.write(chunk)
                    pbar.update(size)

    def forward(self, write_path: str):
        response = requests.get(self.url, stream=True)
        response.raise_for_status()
        self.download(response=response, write_path=write_path)

    def __call__(self, figshare_id: str, write_path: str):

        self.figshare_id = figshare_id

        return self.forward(write_path=write_path)


# -- function: ----------------------------------------------------------------
def figshare_downloader(
    figshare_id: str,
    write_path: str,
    chunk_size: int = int(8 * 1024 * 1024),
):
    """
    Download a file from Figshare.

    Args:
        figshare_id: str
            The ID of the file to download.
        write_path: str
            The path to write the file to.
        url_prefix: Literal["files", "articles"]
            The prefix of the URL to download the file from.
        chunk_size: int
            The size of the chunk to download the file in.
    """
    downloader = FigshareDownloader(chunk_size=chunk_size)
    return downloader(figshare_id=figshare_id, write_path=write_path)



from ._human_hematopoiesis import human_hematopoiesis
from ._pancreatic_endocrinogenesis import pancreatic_endocrinogenesis
from ._larry_in_vitro import larry

__all__ = ["pancreatic_endocrinogenesis", "larry", "human_hematopoiesis"]



# -- import packages: ---------------------------------------------------------
import ABCParse
import pathlib
import yaml

# -- set types: ---------------------------------------------------------------
from typing import Any, Dict, Union


# -- operational class: -------------------------------------------------------
class HParams(ABCParse.ABCParse):
    """scDiffEq container for HyperParams

    Attributes:
        _yaml_path (Union[pathlib.Path, str]): Path to the hparams file created by Lightning.
        _yaml_file (dict): Dictionary containing the contents of the yaml file.
        _attrs (Dict[str, Any]): Formatted attribute dictionary from hparams.yaml.
    """

    def __init__(self, yaml_path: Union[pathlib.Path, str]) -> None:
        """Initialize the HParams object by providing a path to thecorresponding yaml file (created by Lightning)

        Args:
            yaml_path (Union[pathlib.Path, str]): Path to the hparams file created by Lightning.

        Returns:
            None
        """
        self.__configure__(locals())

    def _read(self) -> None:
        """Read path to yaml file and set as class attribute

        Returns:
            None
        """
        if not hasattr(self, "_file"):
            self._yaml_file = yaml.load(open(self._yaml_path), Loader=yaml.Loader)

    def __configure__(self, kwargs, private=["yaml_path"]) -> None:
        """Set hparams as class attributes

        Args:
            kwargs (dict): Dictionary of keyword arguments.
            private (list, optional): List of private attributes, by default ["yaml_path"]

        Returns:
            None
        """
        self.__parse__(kwargs, private=private)
        self._read()
        for key, val in self._yaml_file.items():
            setattr(self, key, val)

    @property
    def _ATTRS(self) -> Dict[str, Any]:
        """Formatted attribute dictionary from hparams.yaml

        Returns:
            Dict[str, Any]
            Dictionary of attributes.
        """
        self._attrs = {
            attr: getattr(self, attr)
            for attr in self.__dir__()
            if not attr[0] in ["_", "a"]
        }
        return self._attrs

    def __getitem__(self, attr: str) -> Any:
        """Format version key and return path

        Args:
            attr (str): Attribute name.

        Returns:
            Any
            Attribute value.
        """
        return self._ATTRS[attr]

    def __repr__(self) -> str:
        """Return a readable representation of the discovered hyperparameters

        Returns:
            str
            Readable representation of the hyperparameters.
        """
        string = "HyperParameters\n"
        for attr, val in self._ATTRS.items():
            string += "\n  {:<34}: {}".format(attr, val)

        return string

    def __call__(self) -> Dict[str, Any]:
        """Return formatted dictionary of attributes from the hparams.yaml

        Returns:
            Dict[str, Any]
            Dictionary of attributes.
        """
        return self._ATTRS



# -- import packages: ---------------------------------------------------------
import ABCParse
import adata_query
import anndata
import autodevice
import lightning
import numpy as np
import pandas as pd
import torch

# -- set type hints: ----------------------------------------------------------
from typing import Optional

# -- operational cls: ---------------------------------------------------------
class Simulation(ABCParse.ABCParse):
    """Sampled trajectories from an scDiffEq model"""

    def __init__(
        self,
        use_key: str = "X_pca",
        time_key: str = "Time point",
        N: int = 1,
        device: Optional[torch.device] = autodevice.AutoDevice(),
        *args,
        **kwargs,
    ):
        """
        Args:
            use_key (str): description. **Default**: "X_pca"

            N (int): number of trajectories to sample from the model. **Default**: 2000

            device (bool): description. **Default**: "cuda:0"

        Returns:
            None
        """

        self.__parse__(locals())

        self._T_GIVEN = False

    @property
    def _adata_init(self): ...

    @property
    def idx(self):
        """ """
        if not hasattr(self, "_idx"):
            self._idx = self._adata.obs.index
        return self._idx

    @property
    def Z0(self) -> torch.Tensor:
        """ """
        if not hasattr(self, "_Z0"):
            self._Z0 = adata_query.fetch(
                self._adata[self.idx],
                key=self._use_key,
                torch=True,
                device=self._device,
            )
            if self._N > 1:
                self._Z0 = self._Z0[None, :, :]
                self._Z0 = self._Z0.expand(self._N, -1, -1).flatten(0, 1)
        return self._Z0

    @property
    def _TIME(self) -> pd.Series:
        if not self._T_GIVEN:
            return self._adata.obs[self._time_key]
        return self.t

    @property
    def _T_MIN(self) -> float:
        """ """
        return self._TIME.min()

    @property
    def _T_MAX(self) -> float:
        """ """
        return self._TIME.max()

    @property
    def _N_STEPS(self) -> float:
        return int(((self._T_MAX - self._T_MIN) / self._dt) + 1)

    @property
    def t(self) -> torch.Tensor:
        """ """
        if not hasattr(self, "_t"):
            self._t = torch.linspace(
                self._T_MIN,
                self._T_MAX,
                self._N_STEPS,
            ).to(self._device)
        return self._t

    @property
    def _N_CELLS(self):
        return self._N * len(self.idx)

    def forward(self, Z0, t) -> torch.Tensor:
        return self._diffeq.forward(Z0, t).detach().cpu().flatten(0, 1).numpy()

    def _to_adata_sim(self, Z_hat: np.ndarray) -> anndata.AnnData:
        """
        Args:
            Z_hat (np.ndarray)

        Returns:
            adata_sim (anndata.AnnData)
        """
        adata_sim = anndata.AnnData(Z_hat)
        adata_sim.obs["t"] = np.repeat(self.t.detach().cpu().numpy(), self._N_CELLS)
        adata_sim.obs["z0_idx"] = np.tile(np.tile(self.idx, self._N_STEPS), self._N)
        adata_sim.obs["sim_i"] = np.tile(
            np.arange(self._N).repeat(len(self.idx)), self._N_STEPS
        )
        adata_sim.obs["sim"] = adata_sim.obs["z0_idx"].astype(str) + adata_sim.obs[
            "sim_i"
        ].astype(str)
        adata_sim.uns["sim_idx"] = self.idx
        adata_sim.uns["simulated"] = True

        return adata_sim

    def __call__(
        self,
        diffeq,
        adata: anndata.AnnData,
        idx: pd.Index,
        dt: float = 0.1,
        t: Optional[torch.Tensor] = None,
        *args,
        **kwargs,
    ) -> anndata.AnnData:
        """Simulate trajectories by sampling from an scDiffEq model.

        Args:
            diffeq (): lightning model.

            adata (anndata.AnnData): Input AnnDat object.

            idx (pd.Index): cell indices (corresponding to `adata` from which the model should
                initiate sampled trajectories.

        Returns:
            adata_sim (anndata.AnnData)
        """

        self.__update__(locals())

        self._diffeq.to(self._device)

        if not t is None:
            self._T_GIVEN = True

        Z_hat = self.forward(self.Z0, self.t)
        return self._to_adata_sim(Z_hat)


def simulate(
    adata: anndata.AnnData,
    diffeq: lightning.LightningModule,
    idx: Optional[pd.Index] = None,
    use_key: str = "X_pca",
    time_key: str = "Time point",
    N: Optional[int] = 1,
    t: Optional[torch.Tensor] = None,
    dt: Optional[float] = 0.1,
    device: Optional[torch.device] = autodevice.AutoDevice(),
    *args,
    **kwargs,
) -> anndata.AnnData:
    """
    Simulate trajectories by sampling from an scDiffEq model.

    Parameters
    ----------
    adata : anndata.AnnData
        Input AnnData object.
    idx : pd.Index
        Cell indices (corresponding to `adata`) from which the model should initiate sampled trajectories.
    diffeq : lightning.LightningModule
        The differential equation model.
    use_key : str, optional
        adata accession key for the input data. Default is "X_pca".
    N : int, optional
        Number of trajectories to sample from the model. Default is 2000.
    device : torch.device, optional
        Device to run the simulation on. Default is True.

    Returns
    -------
    anndata.AnnData
        AnnData object encapsulating scDiffEq model simulation.
    """
    if diffeq.__repr__() == "scDiffEq":
        diffeq = diffeq.DiffEq

    simulation = Simulation(
        use_key=use_key,
        time_key=time_key,
        N=N,
        device=device,
    )
    return simulation(diffeq=diffeq, adata=adata, idx=idx, t=t, dt=dt)



# -- import packages: ---------------------------------------------------------
import lightning
import logging
import os
import pickle

# -- configure logger: --------------------------------------------------------
logger = logging.getLogger(__name__)


# -- callback cls: ------------------------------------------------------------
class IntermittentSaves(lightning.pytorch.callbacks.Callback):
    def __init__(self, frequency: int = 50) -> None:
        """
        Args:
            frequency (int): save frequency
        """

        super().__init__()

        self.frequency = frequency

    def on_train_batch_end(
        self, trainer, model, outputs, batch, batch_idx, *args, **kwargs
    ):
        """
        Args:
            trainer (lightning.pytorch.Trainer): trainer
            model (lightning.pytorch.LightningModule): model
            outputs (dict): outputs
            batch (torch.Tensor): batch
            batch_idx (int): batch index
        """

        epoch = model.current_epoch

        if (epoch % self.frequency == 0) and (batch_idx == 0):
            dpath = os.path.join(trainer.logger.log_dir, "ckpt_outputs")
            if not os.path.exists(dpath):
                os.mkdir(dpath)
                logger.info(f"Created directory: {dpath}")

            fpath = os.path.join(
                dpath, "epoch_{}.batch_{}.pkl".format(epoch, batch_idx)
            )
            pickle.dump(
                obj=outputs,
                file=open(fpath, "wb"),
                protocol=pickle.HIGHEST_PROTOCOL,
            )



# -- import packages: ---------------------------------------------------------
import lightning
import torch

# -- callback cls: ------------------------------------------------------------
class StochasticWeightAveraging(lightning.pytorch.Callback):
    """
    Stochastic Weight Averaging callback.
    """
    def __init__(self, swa_lrs: float) -> None:
        """Initialize the callback.

        Args:
            swa_lrs (float): Learning rate for SWA.
        
        Returns:
            None
        """
        super().__init__()
        self.swa_lrs = swa_lrs
        self.swa_state = {}
        self.n_averaged = 0
    
    def on_fit_start(self, trainer, pl_module):
        # Initialize the dictionary to store parameter sums
        for name, param in pl_module.named_parameters():
            if param.requires_grad:
                self.swa_state[name] = torch.zeros_like(param.data)
    
    def on_train_epoch_end(self, trainer, pl_module):
        # Update the running average of parameters
        for name, param in pl_module.named_parameters():
            if param.requires_grad and name in self.swa_state:
                self.swa_state[name] += param.data
        self.n_averaged += 1
    
    def on_train_end(self, trainer, pl_module):
        # Transfer averaged parameters at the end of training
        if self.n_averaged > 0:
            for name, param in pl_module.named_parameters():
                if param.requires_grad and name in self.swa_state:
                    param.data.copy_(self.swa_state[name] / self.n_averaged)
            print(f"SWA completed with {self.n_averaged} models averaged.")



# -- import packages: ---------------------------------------------------------
import neural_diffeqs
import lightning
import torchsde


# -- set typing: --------------------------------------------------------------
from typing import Union, List


# -- MixIn class: -------------------------------------------------------------
class SDEMixIn(lightning.LightningModule):
    """"""
    def __init__(self, *args, **kwargs):
        super().__init__()

    def _configure_SDE(
        self,
        state_size: int,
        mu_hidden: Union[List[int], int] = [400, 400, 400],
        sigma_hidden: Union[List[int], int] = [50, 50, 50],
        mu_activation: Union[str, List[str]] = "LeakyReLU",
        sigma_activation: Union[str, List[str]] = "LeakyReLU",
        mu_dropout: Union[float, List[float]] = 0.2,
        sigma_dropout: Union[float, List[float]] = 0.2,
        mu_bias: bool = True,
        sigma_bias: List[bool] = True,
        mu_output_bias: bool = True,
        sigma_output_bias: bool = True,
        mu_n_augment: int = 0,
        sigma_n_augment: int = 0,
        sde_type="ito",
        noise_type="general",
        brownian_dim=1,
        coef_drift: float = 1.0,
        coef_diffusion: float = 1.0,
    ):
        self.func = neural_diffeqs.NeuralSDE(
            state_size=state_size,
            mu_hidden=mu_hidden,
            sigma_hidden=sigma_hidden,
            mu_activation=mu_activation,
            sigma_activation=sigma_activation,
            mu_dropout=mu_dropout,
            sigma_dropout=sigma_dropout,
            mu_bias=mu_bias,
            sigma_bias=sigma_bias,
            mu_output_bias=mu_output_bias,
            sigma_output_bias=sigma_output_bias,
            mu_n_augment=mu_n_augment,
            sigma_n_augment=sigma_n_augment,
            sde_type=sde_type,
            noise_type=noise_type,
            brownian_dim=brownian_dim,
            coef_drift=coef_drift,
            coef_diffusion=coef_diffusion,
        )
        
    def _configure_integrator(self, adjoint=False):
        if adjoint:
            self._integrator = torchsde.sdeint_adjoint
        else:
            self._integrator = torchsde.sdeint
            
    def __repr__(self):
        return "scDiffEq MixIn: SDEMixIn"




import torch


class PotentialMixIn(object):
    """Mixin class for potential-based operations.

    This class provides methods for potential-based operations,
    including validation step with gradient computation enabled.

    Attributes:
        None

    Methods:
        __init__: Initialize the PotentialMixIn object.
        validation_step: Perform a validation step with gradient computation enabled.

    """
    def __init__(self):
        """Initialize the PotentialMixIn object.

        Args:
            None

        Returns:
            None

        """
        super().__init__()

    @torch.enable_grad()
    @torch.inference_mode(False)
    def validation_step(self, batch, batch_idx, *args, **kwargs):
        """Perform a validation step with gradient computation enabled.

        This method performs a validation step with gradient computation enabled,
        if the model is not in pre-training mode.

        Args:
            batch: Input batch of data.
            batch_idx (int): Index of the batch.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            None

        """
        if not self.PRETRAIN:
            return self.step(batch, batch_idx, stage="validation")



import pandas as pd
import re

def filter_df(df, regex, include=[True], axis=1):
    """
    Filters a pandas dataframe using regular expressions based on the provided words.

    Parameters:
    df (pd.DataFrame): The input dataframe to be filtered.
    words (list): A list of words to be included or excluded in the filter.
    include (bool): If True, include only columns/rows that match the words; if False, exclude columns/rows that match the words. Default is True.
    axis (int): The axis along which to filter. 0 for rows, 1 for columns. Default is 1 (columns).

    Returns:
    pd.DataFrame: The filtered dataframe.
    """
    
    
    _df = df.copy()
    
    
    for regex_pattern, include_pattern in zip(regex, include):
        # If exclude is chosen, modify regex pattern to exclude the words
        if not include_pattern:            
            # Join words with '|' as a regex pattern to match any of the words
            # regex_pattern = '|'.join(regex_pattern) # for list of lists - unlikely use-case.
            # only works if not a list... 
            # can think about how to include this or if it's even necessary sometime later on....
            
            regex_pattern = f"^(?!.*(?:{regex_pattern})).*$"
        
        # Filter the dataframe using the regex pattern
        _df = _df.filter(regex=regex_pattern, axis=axis)
    return _df

def _filter_df(df, words, include=True, axis=1):
    """
    Filters a pandas dataframe using regular expressions based on the provided words.

    Parameters:
    df (pd.DataFrame): The input dataframe to be filtered.
    words (list): A list of words to be included or excluded in the filter.
    include (bool): If True, include only columns/rows that match the words; if False, exclude columns/rows that match the words. Default is True.
    axis (int): The axis along which to filter. 0 for rows, 1 for columns. Default is 1 (columns).

    Returns:
    pd.DataFrame: The filtered dataframe.
    """

    # Join words with '|' as a regex pattern to match any of the words
    regex_pattern = '|'.join(words)

    # If exclude is chosen, modify regex pattern to exclude the words
    if not include:
        regex_pattern = f"^(?!.*(?:{regex_pattern})).*$"

    # Filter the dataframe using the regex pattern
    filtered_df = df.filter(regex=regex_pattern, axis=axis)

    return filtered_df



# -- import packages: ----------------------------------------------------------
import sklearn
import pandas as pd
import numpy as np
import ABCParse


from ...core import utils

# -- set typing: ---------------------------------------------------------------
NoneType = type(None)



# -- Operational class: --------------------------------------------------------
class NeighborSmoothing(ABCParse.ABCParse):
    _FIT = False

    def __init__(
        self, mode: [NoneType, str] = "radius", radius: int = 1, neighbors: int = 15
    ):
        """
        Parameters:
        -----------
        mode
            type: [NoneType, str]
            default: "radius"

        radius
            type: int
            default: 1

        neighbors
            type: int
            default: 15

        Returns:
        --------
        None

        Notes:
        ------

        """

        self.__parse__(locals(), public=[None])
        self._AVAILABLE_MODELS = {
            "radius": sklearn.neighbors.RadiusNeighborsRegressor,
            "neighbors": sklearn.neighbors.KNeighborsRegressor,
        }

        self._AVAILABLE_MODEL_ARGS = {
            "radius": self._radius,
            "neighbors": self._neighbors,
        }

    def _configure_neighbor_model(self):
        """
        Sets the model and the supplied argument(s).
        """
        arg = self._AVAILABLE_MODEL_ARGS[self._mode]
        neighbor_model = self._AVAILABLE_MODELS[self._mode]
        return neighbor_model(arg)

    @property
    def NEIGHBOR_MODEL(self):
        if not hasattr(self, "_neighbor_model"):
            self._neighbor_model = self._configure_neighbor_model()
        return self._neighbor_model

    @property
    def _DO_SMOOTHING(self):
        return not isinstance(self._mode, NoneType)

    def fit(self, X_umap: np.ndarray, value: np.ndarray) -> None:
        self.NEIGHBOR_MODEL.fit(X_umap, value)
        self._FIT = True

    def predict(self, X_umap: np.ndarray) -> np.ndarray:
        return self.NEIGHBOR_MODEL.predict(X_umap)


# -- import packages: ---------------------------------------------------------
import ABCParse
import anndata
import numpy as np
import scipy.sparse


# -- Set typing: --------------------------------------------------------------
from typing import Optional


# -- Controller class: --------------------------------------------------------
class DistancesHandler(ABCParse.ABCParse):
    def __init__(self, distances_key: str = "distances", *args, **kwargs):
        """
        Args:
            distances_key (str)

        """
        self.__parse__(locals())

    @property
    def distances(self) -> scipy.sparse.csr_matrix:
        if not hasattr(self, "_distances"):
            D = self._adata.obsp[self._distances_key].copy()
            D.data += 1e-6
            self._distances = D
        return self._distances

    @property
    def n_counts(self):
        return (self.distances > 0).sum(axis=1).A.flatten()

    @property
    def n_neighbors(self):
        if not hasattr(self, "_n_neighbors") or self._n_neighbors is None:
            return self.n_counts.min()
        if self.n_counts.min() == 0:
            return self._n_neighbors
        return min(self.n_counts.min(), self._n_neighbors)

    @property
    def rows(self):
        return np.where(self.n_counts > self.n_neighbors)[0]

    @property
    def cumsum_neighbors(self):
        return np.insert(self.n_counts.cumsum(), 0, 0)

    @property
    def _data(self):
        if not hasattr(self, "_private_data"):
            self._private_data = self.distances.data
        return self._private_data

    def _idx_filter(self, row):
        """clean up indices of a row if criteria are met before
        passing to this function"""
        n0, n1 = self.cumsum_neighbors[row], self.cumsum_neighbors[row + 1]
        rm_idx = self._data[n0:n1].argsort()[self.n_neighbors :]
        self._data[rm_idx] = 0

    def _return_mode_distances(self):
        """scvelo also enables return mode on connectivities. we'll
        forego that for now."""
        # Create a new CSR matrix with exactly n_neighbors per row
        n_obs = self._adata.n_obs
        indices = []
        
        for i in range(n_obs):
            # Get row indices
            row_indices = self.distances.indices[self.distances.indptr[i]:self.distances.indptr[i+1]]
            # Sort by distance (closest first)
            row_data = self.distances.data[self.distances.indptr[i]:self.distances.indptr[i+1]]
            idx_sorted = np.argsort(row_data)
            
            # Take only the n_neighbors closest neighbors
            if len(idx_sorted) >= self.n_neighbors:
                nn_idx = row_indices[idx_sorted[:self.n_neighbors]]
            else:
                # If not enough neighbors, pad with the cell's own index
                nn_idx = np.zeros(self.n_neighbors, dtype=int)
                nn_idx[:len(idx_sorted)] = row_indices[idx_sorted]
                nn_idx[len(idx_sorted):] = i
            
            indices.append(nn_idx)
            
        return np.array(indices)

    def __call__(
        self, adata: anndata.AnnData, n_neighbors: Optional[int] = None, *args, **kwargs
    ):
        """Return distance matrix indices

        Args:
            adata (anndata.AnnData)

            n_neighbors (Optional[int])
        """
        self.__update__(locals())

        _ = [self._idx_filter(row) for row in self.rows]
        self.distances.eliminate_zeros()
        self.distances.data -= 1e-6

        return self._return_mode_distances()


# -- API-facing function: -----------------------------------------------------
def get_neighbor_indices(
    adata: anndata.AnnData,
    n_neighbors: Optional[int] = None,
    distances_key: Optional[str] = "distances",
):
    """Return distance matrix indices

    Args:
        adata (anndata.AnnData): anndata.

        n_neighbors (Optional[int]): 
        
        distances_key (Optional[str]): Key accessor to cell neighbor distances
        in ``adata.obsp``. **Default**: "distances".

    Returns:
        (np.ndarray)

    """

    distances_handler = DistancesHandler(distances_key=distances_key)
    return distances_handler(adata=adata, n_neighbors=n_neighbors)




# -- import packages: ---------------------------------------------------------
import ABCParse
import pandas as pd


# -- set typing: --------------------------------------------------------------
from typing import Optional


# -- operational class: -------------------------------------------------------
class GroupedMetrics(ABCParse.ABCParse):
    """Container cls for grouping and reporting metric_df by epochs"""

    def __init__(
        self,
        groupby: Optional[str] = "epoch",
        *args,
        **kwargs,
    ) -> None:
        
        """"""
        self.__parse__(locals())

    def _aggr(self, df, regex: str = "training"):
        """ """
        return df.filter(regex=regex).dropna().sum(1).mean()

    @property
    def _GROUPED(self):
        """ """
        return self._metrics_df.groupby(self._groupby)

    def forward(self, stage: str = "training") -> pd.Series:
        """ """
        return self._GROUPED.apply(self._aggr, regex=stage).dropna()

    @property
    def training(self) -> pd.Series:
        """ """
        return self.forward(stage="training")

    @property
    def validation(self) -> pd.Series:
        """ """
        return self.forward(stage="validation")

    @property
    def df(self) -> pd.DataFrame:
        """ """
        df = pd.concat([self.training, self.validation], axis=1).dropna()
        df.columns = ["training", "validation"]
        return df

    @property
    def lr(self):
        return self._GROUPED["opt_param_group_lr"].mean()

    def __call__(self, metrics_df: pd.DataFrame, *args, **kwargs) -> pd.DataFrame:
        """

        Args:
            metrics_df (pd.DataFrame)
            
        Returns:
            df (pd.DataFrame): grouped (by epoch) metrics_df.
        """
        self.__update__(locals())

        return self.df



# -- import packages: ---------------------------------------------------------
import ABCParse
import anndata
import logging
import os
import pathlib

# -- import local dependencies: -----------------------------------------------
from ._figshare_downloader import figshare_downloader

# -- configure logger: --------------------------------------------------------
logger = logging.getLogger(__name__)


# -- preprocessing dataset cls: -----------------------------------------------
class UnProcessedPancreaticEndocrinogenesisDataset(ABCParse.ABCParse):
    FIGSHARE_ID = "54151331"
    RAW_FNAME = "_downloaded.pancreas.h5ad"
    PROCESSED_FNAME = "pancreas.pp.h5ad"

    def __init__(self, data_dir=os.getcwd(), *args, **kwargs):
        self.__parse__(locals())

        if not self.data_dir.exists():
            self.data_dir.mkdir()

    @property
    def _scdiffeq_parent_data_dir(self):
        path = pathlib.Path(self._data_dir).joinpath("scdiffeq_data")
        if not path.exists():
            path.mkdir()
        return path

    @property
    def data_dir(self):
        path = self._scdiffeq_parent_data_dir.joinpath("pancreatic_endocrinogenesis")
        if not path.exists():
            path.mkdir()
        return path

    @property
    def raw_h5ad_path(self) -> pathlib.Path:
        return self.data_dir.joinpath(self.RAW_FNAME)

    @property
    def processed_h5ad_path(self) -> pathlib.Path:
        return self.data_dir.joinpath(self.PROCESSED_FNAME)

    def download_raw(self):
        figshare_downloader(
            figshare_id=self.FIGSHARE_ID,
            write_path=str(self.raw_h5ad_path),
            url_prefix="files",
        )

    @property
    def raw_adata(self):
        if not hasattr(self, "_raw_adata"):
            if not self.raw_h5ad_path.exists():
                self.download_raw()

            raw_adata = anndata.read_h5ad(self.raw_h5ad_path)
            self._raw_idx = raw_adata.obs.index

            del raw_adata.uns["pca"]
            del raw_adata.uns["neighbors"]
            del raw_adata.obsm["X_pca"]
            del raw_adata.obsm["X_umap"]
            del raw_adata.obsp["distances"]
            del raw_adata.obsp["connectivities"]

            raw_adata.layers["X_counts"] = raw_adata.X

            raw_adata.obs = raw_adata.obs.reset_index()
            raw_adata.obs.index = raw_adata.obs.index.astype(str)

            self._raw_adata = raw_adata

        return self._raw_adata

    @property
    def pp_adata(self):
        if not hasattr(self, "_pp_adata"):
            if not self.processed_h5ad_path.exists():
                self._pp_adata = self.get()
            else:
                self._pp_adata = anndata.read_h5ad(self.processed_h5ad_path)
        return self._pp_adata

    def pp(self, adata, min_genes=200, min_cells=3):

        from .. import scanpy as sc

        adata = adata.copy()

        sc.pp.filter_cells(adata, min_genes=min_genes)
        sc.pp.filter_genes(adata, min_cells=min_cells)
        sc.pp.log1p(adata)
        sc.pp.highly_variable_genes(adata)
        adata = adata[:, adata.var["highly_variable"]].copy()

        return adata

    def dimension_reduction(self, adata: anndata.AnnData):

        from .. import io
        import sklearn
        import umap

        adata = adata.copy()

        self.SCALER_MODEL = sklearn.preprocessing.StandardScaler()
        self.PCA_MODEL = sklearn.decomposition.PCA(n_components=50)
        self.UMAP_MODEL = umap.UMAP(
            n_components=2, n_neighbors=30, random_state=0, min_dist=0.5
        )

        adata.obsm["X_scaled"] = self.SCALER_MODEL.fit_transform(adata.X.toarray())
        adata.obsm["X_pca"] = self.PCA_MODEL.fit_transform(adata.obsm["X_scaled"])
        adata.obsm["X_umap"] = self.UMAP_MODEL.fit_transform(adata.obsm["X_pca"])

        io.write_pickle(
            self.SCALER_MODEL,
            self.data_dir.joinpath("pancreatic_endocrinogenesis.pp.scaler_model.pkl"),
        )
        io.write_pickle(
            self.PCA_MODEL,
            self.data_dir.joinpath("pancreatic_endocrinogenesis.pp.pca_model.pkl"),
        )
        io.write_pickle(
            self.UMAP_MODEL,
            self.data_dir.joinpath("pancreatic_endocrinogenesis.pp.umap_model.pkl"),
        )

        return adata

    def annotate_time(self, adata):

        idx = adata.obs.index
        t0_str_idx = adata.obs[adata.obs["clusters"] == "Ductal"].index
        t0_idx = idx.isin(t0_str_idx)
        t1_idx = ~idx.isin(t0_str_idx)
        t = np.zeros(len(idx))
        t[t1_idx] = 1
        adata.obs["t"] = t

        return adata

    def get(self):

        adata_pp = self.pp(self.raw_adata)
        adata_pp = self.dimension_reduction(adata_pp)
        adata = self.annotate_time(adata_pp)
        adata.write_h5ad(self.processed_h5ad_path)
        return adata


def _unprocessed_pancreas(data_dir: str = os.getcwd()) -> anndata.AnnData:
    """Pancreas dataset

    Parameters
    ----------
    data_dir : str, default=os.getcwd()
        Path to the directory where the data will be saved.

    Returns
    -------
    anndata.AnnData
        Processed AnnData object.
    """
    data_handler = UnProcessedPancreaticEndocrinogenesisDataset(data_dir=data_dir)
    return data_handler.pp_adata


# -- dataset cls: -------------------------------------------------------------
class PancreaticEndocrinogenesisDataset(ABCParse.ABCParse):
    _FNAME = "pancreatic_endocrinogenesis"
    _ADATA_FNAME = f"adata.{_FNAME}.cytotrace.h5ad"
    figshare_ids = {
        "adata": {"file_id": 54151208, "fname": _ADATA_FNAME},
        "scaler": {"file_id": 54151202, "fname": f"{_FNAME}.scaler.pkl"},
        "pca": {"file_id": 54151205, "fname": f"{_FNAME}.pca.pkl"},
        "umap": {"file_id": 54151199, "fname": f"{_FNAME}.umap.pkl"},
    }

    def __init__(
        self,
        data_dir=os.getcwd(),
        force_download: bool = False,
        *args,
        **kwargs,
    ):

        self.__parse__(locals())

    @property
    def _scdiffeq_parent_data_dir(self):
        path = pathlib.Path(self._data_dir).joinpath("scdiffeq_data")
        path.mkdir(exist_ok=True, parents=True)
        return path

    @property
    def data_dir(self):
        path = self._scdiffeq_parent_data_dir.joinpath("pancreatic_endocrinogenesis")
        path.mkdir(exist_ok=True, parents=True)
        return path

    @property
    def h5ad_path(self) -> pathlib.Path:
        return self.data_dir.joinpath(self._ADATA_FNAME)

    def download(self):
        for _, val in self.figshare_ids.items():
            figshare_downloader(
                figshare_id=val["file_id"],
                write_path=self.data_dir.joinpath(val["fname"]),
            )

    @property
    def adata(self) -> anndata.AnnData:
        if not hasattr(self, "_adata"):
            if not self.h5ad_path.exists() or self._force_download:
                self.download()
            self._adata = anndata.read_h5ad(self.h5ad_path)
        return self._adata


def pancreatic_endocrinogenesis(
    data_dir: str = os.getcwd(),
    filter_genes: bool = True,
    reduce_dimensions: bool = True,
    force_download: bool = False,
    download_unprocessed: bool = False,
) -> anndata.AnnData:
    """Pancreas dataset

    Args:
        data_dir: str, default=os.getcwd()
            Path to the directory where the data will be saved.
        filter_genes: bool, default=True
            Whether to filter genes.
        reduce_dimensions: bool, default=True
            Whether to reduce dimensions.
        force_download: bool, default=False
            Whether to force download the data.
        download_unprocessed: bool, default=False
            Whether to download the unprocessed data.

    Returns:
        anndata.AnnData: Preprocessed AnnData object.
    """
    if download_unprocessed:
        return _unprocessed_pancreas(
            data_dir=data_dir,
            filter_genes=filter_genes,
            reduce_dimensions=reduce_dimensions,
            force_download=force_download,
        )
    else:
        data_handler = PancreaticEndocrinogenesisDataset(
            data_dir=data_dir,
            force_download=force_download,
        )
        return data_handler.adata
